////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "XmdvTool.H"
#include "macros.H"
#include "ogl.H"
#include "okcdata.H"
#include "parcoord.H"
#include "scatter.H"
#include "glyph.H"
#include "dimstack.H"
#include "cluster.H"
#include "cluster_parcoord.H"
//jing
#include "cluster_scatter.H"
#include "cluster_glyph.H"
#include "cluster_dimstack.H"
#include "cluster_nav.H"
#include "cluster_hist.H"
#include "brush.H"
#include "brush_oper.H"
#include "magnify.H"
#include "glyphtool.H"
#include "stacktool.H"
#include "parcoordtool.H"
#include "treemap.H"
#include "multi_select.H"
#include <iostream.h> 
//jing
//#include <unistd.h>
//jing
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include "stats.H"
#include "X11/keysym.h"

#ifdef WIN32
#define strcasecmp stricmp
#endif

int DSIZE = 1;
struct okcdata *dataset = 0;
unsigned long interface_color[NUM_INTERFACE_ITEM];
char oldType[50];

static char interface_file[200];

static void appInitGlobals()
{
    dataset = 0;
    interface_color[BG] = 0xffffff;
    interface_color[TEXT] = 0x000000;
    interface_color[DATA] = 0x61f369;
    interface_color[HIGHLIGHT] = 0xf20000;
    interface_color[OVERLAP] = 0x000000;
    interface_color[PAINT] = 0x980098;
    interface_color[GRID1] = 0xff0f0f;
    interface_color[GRID2] = 0xafafaf;
    interface_color[GRID3] = 0xeaeaea;
    interface_color[STAT1] = 0x0000ea;
    interface_color[STAT2] = 0xc30000;
    interface_color[BRUSH1] = 0xafafea;
    interface_color[BRUSH2] = 0xafafaf;
    interface_color[BRUSH3] = 0xafaf75;
    interface_color[BRUSH4] = 0xafaf3a;
    interface_color[SBB] = 0xff7138;
}

static int appSetupDataValuesDialog(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) return TCL_OK;

    Tcl_VarEval(interp, "[W DataValuesText] conf -state normal",0);
    Tcl_VarEval(interp, "[W DataValuesText] delete 0.0 end",0);

    char buf[500];
    char str[1024];
    int i;
    strcpy(str,"Oper: ");
    for (i=0;i<dataset->dims;i++) {
        strcpy(buf,dataset->names[i]);
        buf[12] = '\0';
        sprintf(str,"%s %12s",str,buf);
    }
    strcat(str,"\n");
    Tcl_VarEval(interp, "[W DataValuesText] insert end \"",str,"\"",0);

//    char sep[strlen(str)+1];
//jing
    char* sep = new char[strlen(str)+1];
//jing
    for (i=0;i<(int)strlen(str);i++) sep[i]='-';
    sep[strlen(str)-1] = '\0';
    strcat(sep,"\n");
    Tcl_VarEval(interp, "[W DataValuesText] insert end \"",sep,"\"",0);

    int count = 0;
    for (int op=0; op<num_oper; op++)
    {
        // check if the operation has OPER_VALUES set
        if (!(brush_ops[op].oper & VALUES)) continue;

        ARR<float> data(dataset->dims);
        for (int i=0; i<dataset->data_size; i++)
        {
            dataset->GetData(data, i);

            if (BrushOperCheckPoint(data, op))
            {
                sprintf(str, "%4d: ", op+1);
                for (int j=0; j<dataset->dims; j++)
                {
                    sprintf(str, "%s %12.3f",str,data[j]);
                }
                strcat(str,"\n");
                Tcl_VarEval(interp, "[W DataValuesText] insert end \"",str,"\"",0);
                count++;
            }
        }

        // Check if we should add an average too
        if (brush_ops[op].oper & AVERAGE)
        {
            if (BrushOperCompAverage(op, data)==0) continue;

            sprintf(str, "Avg%d: ", op+1);
            for (int j=0; j<dataset->dims; j++)
            {
                sprintf(str, "%s %12.3f",str,data[j]);
            }
            strcat(str,"\n");
            Tcl_VarEval(interp, "[W DataValuesText] insert end \"",str,"\"",0);
        }
    }

    Tcl_VarEval(interp, "[W DataValuesText] insert end \"",sep,"\"",0);

    sprintf(str,"Total number of brushed data points = %d\n",count);
    Tcl_VarEval(interp, "[W DataValuesText] insert end \"",str,"\"",0);

    Tcl_VarEval(interp, "[W DataValuesText] conf -state disabled",0);
//jing
	delete sep;
//jing
    return TCL_OK;
}

void appRefreshAllCanvas(Tcl_Interp *interp, grCanvas *except_this)
{
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    if (canvas!=except_this) canvas->Refresh();

    grCanvas *nav = grFindCanvas(interp,"structBrushCanvas");
    if (nav && nav!=except_this) nav->Refresh();

    grCanvas *parcoordtool = grFindCanvas(interp,"minimapCanvas");
    if (parcoordtool && parcoordtool!=except_this) parcoordtool->Refresh();

    grCanvas *hist = grFindCanvas(interp,"histoCanvas");
    if (hist && hist!=except_this) hist->Refresh();

    grCanvas *treemap = grFindCanvas(interp,"treemapCanvas");
    if (treemap && treemap!=except_this) treemap->Refresh();

    grCanvas *glyphtool = grFindCanvas(interp,"glyphToolboxCanvas");
    if (glyphtool && glyphtool!=except_this) glyphtool->Refresh();

    grCanvas *stacktool = grFindCanvas(interp,"stackToolboxCanvas");
    if (stacktool && stacktool!=except_this) stacktool->Refresh();

    grCanvas *aux = grFindCanvas(interp,"auxillaryDisplayCanvas");
    if (aux && aux!=except_this) aux->Refresh();

    char str[50];   
	strcpy (str, "wm state [W DataSummaryDialog]");
    Tcl_Eval(interp, str);
    if (strcmp(interp->result,"normal")==0)
        appSetupBrushStatistics(interp);

}

static int appMainCanvasRefresh(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();
    return TCL_OK;
}

static int appRefreshAllCanvas(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    appRefreshAllCanvas(interp);
    return TCL_OK;
}

void appCanvasInputKeyPress(grCanvas *canvas, int key)
{
    if (canvas->shiftkey) {
        // Go into multiselect mode
        multiSelectSetState(canvas,true);
        return;
    } else
    if (canvas->controlkey && curr_brush->bound_type == Brush::RAMP) {
        canvas->SetCursor("right_ptr");
    }

    switch (key) {
        case 'z': case 'Z':
            MagnifyBrushRegion(canvas);
            break;

        case 'a': case 'A':
            Demagnify(canvas);
            break;

        case 'e': case 'E':
            DSIZE++;
            DSIZE = CLAMP(DSIZE,1,10);
            canvas->Refresh();
            break;

        case 'r': case 'R':
            DSIZE--;
            DSIZE = CLAMP(DSIZE,1,10);
            canvas->Refresh();
            break;
        default: break;
    }
}

void appCanvasInputKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        multiSelectSetState(canvas,false);
        return;
    }
    canvas->SetDefaultCursor();
}

static int appSetDoubleBuffer(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    grCanvas::doublebuffer = atoi(argv[1]);
    return TCL_OK;
}

static int appSetCanvasZoom(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    double zoom = (atof(argv[1])-1)/40. + 1;

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    canvas->zoom.X = canvas->zoom.Y = zoom;
    canvas->Refresh();

    return TCL_OK;
}

static int appDependentDimensionHasChanged(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int which = atoi(argv[1]) - 1;
    if (dep_dim == which) return TCL_OK;

    dep_dim = which;
    grCanvas *canvas = grFindCanvas(interp,"treemapCanvas");
    assert(canvas);
    cfTreeMapInit();
    cfTreeMapRedrawCanvas(canvas);

    canvas->Refresh();

    return TCL_OK;
}

void appCanvasBrushHasChanged(Tcl_Interp *interp)
{
    if (dataset) BrushOperMarkDataset();
//    if (btree) hierModifyExtentsByBrushParameters();

}

void appShowCaption(Tcl_Interp *interp, const char *str)
{
//jing
    char stri[50];   
	strcpy(stri, "ClearCaption");
	Tcl_Eval(interp, stri);
//    Tcl_Eval(interp, "ClearCaption");
//jing
    Tcl_VarEval(interp, "ShowCaption \"",str,"\"",0);
}

static int appSetActiveBrush(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int which = atoi(argv[1]);

    SetActiveBrush(which);

    appCanvasBrushHasChanged(interp);
    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static int appSetBrushEnableState(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int which = atoi(argv[1]);
    int state = atoi(argv[2]);

    SetBrushEnableState(which,state);

    appCanvasBrushHasChanged(interp);
    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static int appSetBrushDisplayState(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int which = atoi(argv[1]);
    int state = atoi(argv[2]);

    SetBrushDisplayState(which,state);

    appRefreshAllCanvas(interp);
    return TCL_OK;
}


static int appSetMaskBrushedData(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    maskBrushedData= atoi(argv[1]);

    appRefreshAllCanvas(interp);
    return TCL_OK;
}


static int appSetMaskUnbrushedData(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    maskUnbrushedData= atoi(argv[1]);

    appRefreshAllCanvas(interp);
    return TCL_OK;
}


static int appSetMaskBrush(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    maskBrush= atoi(argv[1]);

    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static int appSetBrushBoundaryType(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int which = atoi(argv[1]);
    int type = atoi(argv[2]);

    SetBrushBoundaryType(which,(type==0)?Brush::STEP:Brush::RAMP);
    appCanvasBrushHasChanged(interp);

    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static int appSaveMainCanvasImage(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *fname = argv[1];

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->WriteImage(fname);

    return TCL_OK;
}

static int appGlobalBrushResize(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *how = argv[1];

    if (strcasecmp(how,"max")==0) {
	curr_brush->Resize(0);
    } else
    if (strcasecmp(how,"half")==0) {
//        curr_brush->Resize(0.5);
        curr_brush->Resize(0);
        curr_brush->Resize(0.5);
    } else
    if (strcasecmp(how,"+")==0) {
        curr_brush->Resize(1.1);
    } else
    if (strcasecmp(how,"-")==0) {
        curr_brush->Resize(0.9);
    }

    appCanvasBrushHasChanged(interp);
    appRefreshAllCanvas(interp);

    return TCL_OK;
}


static int appSetClusterRadius(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset || !btree) return TCL_OK;

    float val = atof(argv[1]);
    if (val==cluster_radius) return TCL_OK;

    cluster_radius = val;

    float root_radius = btree->nodes[0].sqrt_radius;
    hierMarkContourNodes( root_radius*cluster_radius, IS_CONTOUR0 );

    appRefreshAllCanvas(interp);

    return TCL_OK;
}

void appUpdateClusterRadiusSliders(Tcl_Interp *interp)
{
    assert(interp);
    char str[30];
    sprintf(str," %.2f %.2f ",cluster_radius, brushed_radius);
    Tcl_VarEval(interp, "UpdateClusterRadiusSliders ",str,0);
}

static int appSetBrushedRadius(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset || !btree) return TCL_OK;

    float val = atof(argv[1]);
    if (val==brushed_radius) return TCL_OK;

    brushed_radius = val;

    float root_radius = btree->nodes[0].sqrt_radius;
    hierMarkContourNodes( root_radius*brushed_radius, IS_CONTOUR1 );

    appRefreshAllCanvas(interp);

    return TCL_OK;
}


static int appSetBandsOnOff(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    SHOW_BANDS = atoi(argv[1]);

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();

    return TCL_OK;
}

static int appMagnify(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    MagnifyBrushRegion(canvas);

    return TCL_OK;
}

static int appDeMagnify(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    Demagnify(canvas);

    return TCL_OK;
}

static int appSetExtentScalingFactor(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    extent_scaling_fac = atof(argv[1])/100.;
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();

    return TCL_OK;
}

static int appSetDynamicMaskingScale(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    fade_factor = atof(argv[1]);
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    canvas->Refresh();

    return TCL_OK;
}

void appCanvasMousePan(grCanvas *canvas)
{
    Vec2 motion = canvas->MapScreenToData(canvas->curr_pos) - canvas->MapScreenToData(canvas->prev_pos);
    canvas->translate = canvas->translate + motion;
//jing
    char stri[50];   
	strcpy(stri, "CanvasVertScrollCmd sync");
//    Tcl_Eval(Togl_Interp(canvas->GetWidget()),"CanvasVertScrollCmd sync");
//    Tcl_Eval(Togl_Interp(canvas->GetWidget()),"CanvasHorizScrollCmd sync");
    Tcl_Eval(Togl_Interp(canvas->GetWidget()),stri);
	strcpy(stri, "CanvasHorizScrollCmd sync");
    Tcl_Eval(Togl_Interp(canvas->GetWidget()),stri);
//jing
    canvas->Refresh();
}

void appSetupBrushStatistics(Tcl_Interp *interp)
{

    Tcl_Eval(interp,"SetupBrushStats");
}

void appRefreshGlyphPlacementDialog(Tcl_Interp *interp)
{

    Tcl_Eval(interp,"RefreshGlyphPlacement");
}

static int appGetBrushOpString(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int op_num = atoi(argv[1])-1;
    if (op_num >= num_oper) return TCL_OK;
    Tcl_SetResult( interp, (char*) BrushOperGenerateString(atoi(argv[1])-1), TCL_STATIC );
    return TCL_OK;
}

static int appNewBrushOps(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int op_num = BrushOperNew();
    char str[5];
    sprintf(str,"%d",op_num);
    Tcl_SetResult( interp, str, TCL_STATIC );

/*  if (dataset) BrushOperMarkDataset(); // Need to recolor the points
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();
*/
    return TCL_OK;
}

/*
static int appUpdataDisplays(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    appCanvasBrushHasChanged(interp);
    appRefreshAllCanvas(interp);
    return TCL_OK;
}
*/

static int appDeleteBrushOps(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int op_num = atoi(argv[1])-1;
    BrushOperDelete(op_num);

    appCanvasBrushHasChanged(interp);
    appRefreshAllCanvas(interp);

    return TCL_OK;
}

static int appGetBrushOps(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    static char buf[10];
    strcpy(buf, "");
    BrushOperation *op = &brush_ops[atoi(argv[1])-1];

    assert(op);

    // Append the operation(s)
    if (op->oper & HIGH)
        strcat(buf, "1 ");
    else
    if (op->oper & MASK)
        strcat(buf, "0 ");
    else
        strcat(buf, "-1 ");

    if (op->oper & VALUES)
        strcat(buf, "1 ");
    else
        strcat(buf, "0 ");

    if (op->oper & AVERAGE)
        strcat(buf, "1 ");
    else
        strcat(buf, "0 ");

    Tcl_SetResult( interp, buf, TCL_STATIC );
    return TCL_OK;

}

static int appGetNumBrushes(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    BrushOperation *op = &brush_ops[atoi(argv[1])-1];
    assert(op);

    static char brushes[5]="";
    strcpy(brushes,"");
    sprintf(brushes, "%d", op->num_brush);
    Tcl_SetResult( interp, brushes, TCL_STATIC );
    return TCL_OK;
}

static int appGetBrushExpr(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    static char buf[100];
    strcpy(buf, "");
    BrushOperation *op = &brush_ops[atoi(argv[1])-1];
    int idx = atoi(argv[2])-1;
    assert(op);

    // Add a NOT if necessary
    if (op->not[idx])
        strcat(buf, "1 ");
    else
        strcat(buf, "0 ");

    // Add the brush number
    char str[5];
    sprintf(str,"%d ",op->brush[idx]+1);
    strcat(buf, str);

    // Add a conjunction if there are more brushes
    if (idx < op->num_brush-1) {
        switch(op->conj[idx])
        {
        case AND:
            strcat(buf, "1 ");
            break;
        case OR:
            strcat(buf, "2 ");
            break;
        case XOR:
            strcat(buf, "3 ");
            break;
        }
    }

    Tcl_SetResult( interp, buf, TCL_STATIC );
    return TCL_OK;

}

static int appFormBrushExpr(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int cur_op = atoi(argv[1]) - 1;

    brush_ops[cur_op].num_brush= 0;

    brush_ops[cur_op].oper = 0;

    if (strcmp(argv[2], "highlight") == 0)
        brush_ops[cur_op].oper = HIGH;

    if (strcmp(argv[2], "mask") == 0)
        brush_ops[cur_op].oper = MASK;

    if (atoi(argv[3])) { BITSET(brush_ops[cur_op].oper,VALUES); }

    if (atoi(argv[4])) { BITSET(brush_ops[cur_op].oper,AVERAGE); }

    int i;
    for (i=0; i<MAXBRUSH; i++) {
        brush_ops[cur_op].not[i] = atoi(argv[i+5]);
    }

    for (i=0; i<MAXBRUSH; i++) {
        if (atoi(argv[i+9])) brush_ops[cur_op].num_brush++;
        brush_ops[cur_op].brush[i] = atoi(argv[i+9])-1;
    }

    for (i=0; i<MAXBRUSH-1; i++) {
        if (strcmp(argv[i+13],"and") == 0)
            brush_ops[cur_op].conj[i] = AND;

        if (strcmp(argv[i+13],"or") == 0)
            brush_ops[cur_op].conj[i] = OR;

        if (strcmp(argv[i+13],"xor") == 0)
            brush_ops[cur_op].conj[i] = XOR;

    }

    if (brush_ops[cur_op].oper != 0)
        appCanvasBrushHasChanged(interp);

    appRefreshAllCanvas(interp);

    return TCL_OK;
}


static int appLowerOperPriority(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    BrushOperLowerPriority(atoi(argv[1])-1);

    if (dataset) BrushOperMarkDataset(); // Need to recolor the points
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();
    return TCL_OK;
}

static int appRaiseOperPriority(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    BrushOperRaisePriority(atoi(argv[1])-1);

    if (dataset) BrushOperMarkDataset(); // Need to recolor the points
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();
    return TCL_OK;
}

static int appDrawBrushStatsCanvas(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) return TCL_OK;
    int which = atoi(argv[1]);
    char *where = argv[2];

    Tcl_CmdInfo infoPtr;
    if (Tcl_GetCommandInfo(interp,(char*)where,&infoPtr)==0) {
        return 0;
    }
    Tk_Canvas canvas = (Tk_Canvas) infoPtr.clientData;
    assert(canvas);

    statsPlotBrushCoverage(interp,canvas,where,which);

    char str[100];
    sprintf(str,"%f %f",curr_brush->Min(which), curr_brush->Max(which));

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}

static int appGetDatasetSummary(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    char full_path[200];
//    realpath(dataset->datafile, full_path);
//jing
	strcpy(full_path, dataset->datafile);
//jing
    char *short_name = strrchr(full_path, '/');
    *short_name = '\0';
    short_name++;

    struct stat buf;
    stat(dataset->datafile, &buf);

    char str[1000];
    sprintf(str,"%s %s %d %d %d",short_name,full_path, int(buf.st_size), dataset->data_size, dataset->all_dims);

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}

static int appGetDimensionName(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{

    if (!dataset) return TCL_OK;

    int which = atoi(argv[1]);
    if (which < 0 || which >= dataset->all_dims) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    Tcl_SetResult( interp, dataset->names[which], TCL_STATIC );

    return TCL_OK;
}

static int appGetDimensionNames(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    char str[200];
	strcpy(str, dataset->all_names[0]);
    for ( int i = 1; i < dataset->all_dims; i++ )
	{
		strcat(str, " ");
		strcat(str, dataset->all_names[i]);
	}

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}


static int appGetPrincipalNames(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

	int i, prin_num;
	char **prin_names;
	if ( dataset->dims < 3 )
		prin_num = dataset->dims;
	else
		prin_num = 3;
	prin_names = new char *[prin_num];

	char name_num[1];
	for ( i = 0; i < prin_num; i++ )
	{
		prin_names[i] = new char[20];
		strcpy(prin_names[i], "Principal_Component_");
		_itoa( i, name_num, 10);
		strcat(prin_names[i], name_num);
	}

    char str[200];
	strcpy(str, prin_names[0]);
    for ( i = 1; i < prin_num; i++ )
	{
		strcat(str, " ");
		strcat(str, prin_names[i]);
	}

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}

static int appGetFieldInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    int which = atoi(argv[1]);
    if (which < 0 || which >= dataset->all_dims) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    char str[200];
//jing    sprintf(str,"%s %d %f %f",dataset->names[which],dataset->on[which],dataset->dim_min[which],dataset->dim_max[which]);
    sprintf(str,"%s %d %f %f",dataset->all_names[which],dataset->on[which],dataset->all_dim_min[which],dataset->all_dim_max[which]);

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}

static int appGetClusterBrush(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int i; 
	if ( SHOW_BANDS )
		i = 1;
	else
		i = 0;

    char str[200];
	sprintf(str,"%d %f %f %f %f",i,cluster_radius,brushed_radius,fade_factor, 100-extent_scaling_fac*100);

    Tcl_SetResult( interp, str, TCL_STATIC );

    return TCL_OK;
}

static int appSetFieldOnOff(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    int which = atoi(argv[1]);
    int state = atoi(argv[2]);

    dataset->SetFieldOnOff(which,state);
    appRefreshGlyphPlacementDialog(interp);

    if (btree) hierSetMeanOnOff();

    appCanvasBrushHasChanged(interp);

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();

    return TCL_OK;
}

static int appReorderDimension(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    int which = atoi(argv[1]);
    int forward = atoi(argv[2]);

    dataset->ReorderDimension(which, forward);
    if (btree) hierReorderDimension();

	appCanvasBrushHasChanged(interp);

    Tcl_Eval(interp,"SetupDataSummary");
//jing
    Tcl_Eval(interp,"SetupGlyphKeyDialog");

    Tcl_Eval(interp,"SetupDimStackKeyDialog");

    Tcl_Eval(interp,"SetupBrushStats");

    Tcl_Eval(interp,"SetupClusterBrush");
//jing
    Tcl_Eval(interp,"SetupDependentDimensionsMenu");

    appRefreshAllCanvas(interp);
/*

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);
    canvas->Refresh();
*/
    return TCL_OK;
}


static int appGlyphPlacementSort(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    char* name = argv[1];
    int increase = atoi(argv[2]);
    int space = atoi(argv[3]);

	bool success = false;
	int dimnumber;
	for ( int i = 0; i < dataset->all_dims; i++ )
	{
		if ( strcmp(name, dataset->all_names[i]) == 0 )
		{
			success = true;
			dimnumber = i;
			break;
		}
	}
	
	if ( success )
	{
		if ( space )
			dataset->Placement_Sort_Space_Glyph(dimnumber ,increase);
		else
			dataset->Placement_Sort_Glyph(dimnumber ,increase);
	    appRefreshAllCanvas(interp);
	}
    return TCL_OK;
}

static int appGlyphPlacementOriginal(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    int space = atoi(argv[1]);
	if ( space )
		dataset->Placement_Original_Space_Glyph();
	else 
		dataset->Placement_Original_Glyph();

    appRefreshAllCanvas(interp);

    return TCL_OK;
}

static int appGlyphPlacementDataDriven(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

    char* name1 = argv[1];
    char* name2 = argv[2];
    int shift = atoi(argv[3]);

	bool success1 = false;
	bool success2 = false;

	int xdimnumber, ydimnumber;
	for ( int i = 0; i < dataset->all_dims; i++ )
	{
		if ( strcmp(name1, dataset->all_names[i]) == 0 )
		{
			success1 = true;
			xdimnumber = i;
		}
		if ( strcmp(name2, dataset->all_names[i]) == 0 )
		{
			success2 = true;
			ydimnumber = i;
		}
	}
	
	if ( success1 && success2 )
	{
		dataset->Placement_DataDriven_Glyph(xdimnumber ,ydimnumber);
		if ( shift )
		{
			dataset->Placement_Shift_Glyph();
		}
	    appRefreshAllCanvas(interp);
	}

    return TCL_OK;
}


static int appGlyphPlacementDerived(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) {
        Tcl_SetResult( interp, "", TCL_STATIC );
        return TCL_OK;
    }

	int i, prin_num;
	char **prin_names;
	if ( dataset->dims < 3 )
		prin_num = dataset->dims;
	else
		prin_num = 3;
	prin_names = new char *[prin_num];

	char name_num[1];
	for ( i = 0; i < prin_num; i++ )
	{
		prin_names[i] = new char[20];
		strcpy(prin_names[i], "Principal_Component_");
		_itoa( i, name_num, 10);
		strcat(prin_names[i], name_num);
	}
	
	char* name1 = argv[1];
    char* name2 = argv[2];
    int shift = atoi(argv[3]);


	bool success1 = false;
	bool success2 = false;

	int xdimnumber, ydimnumber;
	for ( i = 0; i < prin_num; i++ )
	{
		if ( strcmp(name1, prin_names[i]) == 0 )
		{
			success1 = true;
			xdimnumber = i + dataset->all_dims;
		}
		
		if ( strcmp(name2, prin_names[i]) == 0 )
		{
			success2 = true;
			ydimnumber = i + dataset->all_dims;
		}

	}

	for ( i = 0; i < dataset->all_dims; i++ )
	{
		if ( strcmp(name1, dataset->all_names[i]) == 0 )
		{
			success1 = true;
			xdimnumber = i;
		}
		
		if ( strcmp(name2, dataset->all_names[i]) == 0 )
		{
			success2 = true;
			ydimnumber = i;
		}
	}
	
	if ( success1 && success2 )
	{
		dataset->Placement_Derived_Glyph(xdimnumber ,ydimnumber);
		if ( shift )
		{
			dataset->Placement_Shift_Glyph();
		}
	    appRefreshAllCanvas(interp);
	}

    return TCL_OK;
}

static int appGetCanvasViewParameters(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    Vec2 percent;
    percent.X = 1.0 / canvas->zoom.X;
    percent.Y = 1.0 / canvas->zoom.Y;

    Vec2 org;
    org.X = -canvas->translate.X;
    org.Y = -canvas->translate.Y;

    //if (canvas->origin_at_lower_left) org.Y = 1.0 + canvas->translate.Y - percent.Y;

    char str[100];
    sprintf(str,"%f %f %f %f",org.X,org.Y,percent.X,percent.Y);

    Tcl_SetResult( interp, str, TCL_STATIC );
    return TCL_OK;
}

static int appSetViewScrollTo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    float xorg = atof(argv[1]);
    float yorg = atof(argv[2]);

    grCanvas *canvas = grFindCanvas(interp,"mainCanvas");
    assert(canvas);

    //if (canvas->origin_at_lower_left) yorg = 1.0 - yorg - 1.0/canvas->zoom.Y;

    canvas->translate.X = -xorg;
    canvas->translate.Y = -yorg;

    canvas->Refresh();

    return TCL_OK;
}

static int appQueryGlyphKey(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) return TCL_OK;

    // Compute text strings and place pointers into keylist
    //
    char key_text[500] = "\0";
    for (int i = 0;i < dataset->dims;i++) {
        char str[100];
        sprintf(str, "%s: %.1f degrees\n",dataset->names[i],i*360./dataset->dims);
        strcat(key_text,str);
    }

    Tcl_SetResult( interp, key_text, TCL_STATIC );
    return TCL_OK;
}

static int appQueryDimStackKey(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (!dataset) return TCL_OK;

    char key_text[500] = "\0";
    char str[100];

    // Do horizontal dimensions, then vertical, slow speed to high
    //
    sprintf(str, "Horizontal Dimensions: outer to inner\n");
    strcat(key_text,str);

    int i;
    for (i = 0;i < dataset->dims;i= i+2) {
        sprintf(str, "   %s: cardinality %d\n",dataset->names[i], dataset->cardinality[i]);
        strcat(key_text,str);
    }

    sprintf(str, "Vertical Dimensions: outer to inner\n");
    strcat(key_text,str);

    for (i = 1;i < dataset->dims;i= i+2) {
        sprintf(str, "   %s: cardinality %d\n",dataset->names[i], dataset->cardinality[i]);
        strcat(key_text,str);
    }

    Tcl_SetResult( interp, key_text, TCL_STATIC );
    return TCL_OK;
}

static int appRegisterCanvas(Tcl_Interp *interp, const char *canvas_name, const char *type)
{
    grCanvas *canvas = grFindCanvas(interp,canvas_name);
    assert(canvas);

    canvas->ResetEventHandlers();

    if (strcmp(type,"parcoord")==0) {
        RegisterParcoordCanvas(canvas);
		strcpy(oldType, type);
    } else
    if (strcmp(type,"scatter")==0) {
        RegisterScatterCanvas(canvas);
		strcpy(oldType, type);
    } else
    if (strcmp(type,"glyphs")==0) {
        RegisterGlyphCanvas(canvas);
		strcpy(oldType, type);
    } else
    if (strcmp(type,"dimstack")==0) {
        RegisterDimstackCanvas(canvas);
		strcpy(oldType, type);
    } else
    if (strcmp(type,"hierarchical")==0) {

        if (dataset && btree==0) {
            int N = strlen(dataset->datafile);
//           char cf_file[N+1];
//jing
            char* cf_file = new char[N+1];
//jing
            strcpy(cf_file,dataset->datafile);
            strcpy(&cf_file[N-4],".cf");

            Tcl_VarEval(interp,"LoadClusterFile ",dataset->datafile," ",cf_file,0);
//jing 
			delete cf_file;
//jing
		}
//jing 
      if ( strcmp(oldType, "scatter") == 0 )
		  RegisterHierScatterCanvas(canvas);
      else if ( strcmp(oldType, "glyphs") == 0 )
		  RegisterHierGlyphCanvas(canvas);
      else if ( strcmp(oldType, "dimstack") == 0 )
		  RegisterHierDimstackCanvas(canvas);
	  else
		  RegisterHierParcoordCanvas(canvas);

    } else
    if (strcmp(type,"cfNav")==0) {
        RegisterCfNavCanvas(canvas);
    } else
    if (strcmp(type,"cfHist")==0) {
        RegisterCfHistCanvas(canvas);
    } else
    if (strcmp(type,"minimap")==0) {
        RegisterParcoordToolboxCanvas(canvas);
    } else
    if (strcmp(type,"treemap")==0) {
        RegisterTreemapCanvas(canvas);
    } else
    if (strcmp(type,"glyphtoolbox")==0) {
        RegisterGlyphToolCanvas(canvas);
    } else
    if (strcmp(type,"stacktoolbox")==0) {
        RegisterStackToolCanvas(canvas);

    } else {
        cerr << "### Undefined display mode: \"" << type << "\".\n";
    }


    canvas->Refresh();

    return TCL_OK;
}

static int appRegisterCanvas(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    return appRegisterCanvas(interp,argv[1],argv[2]);
}

static int appModeSelect(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *what = argv[1];

    appRegisterCanvas(interp, "mainCanvas", what);
    appRefreshAllCanvas(interp);

    return TCL_OK;
}

static int appCloseOkcFile(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if (dataset)
	{
		delete dataset->basex;
		delete dataset->basey;
		delete dataset;
	
	}

    dataset = 0;

    hierDeleteClusterTree();

    Tcl_Eval(interp,"SetupDataSummary");
//jing
    Tcl_Eval(interp,"SetupGlyphKeyDialog");
    
    Tcl_Eval(interp,"SetupDimStackKeyDialog");

    //Tcl_Eval(interp,"SetupBrushStats");

//jing
    appRefreshAllCanvas(interp);

    return TCL_OK;
}

static int appLoadOkcFile(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *what = argv[1];
    appCloseOkcFile(clientData,interp,argc,argv);

    dataset = new okcdata;
    dataset->Read(what);

    InitBrushSet();
    InitClusterSet();
    BrushOperListInit();
    BrushOperMarkDataset();

    Tcl_Eval(interp,"SetupDataSummary");
//jing
    Tcl_Eval(interp,"SetupGlyphKeyDialog");

    Tcl_Eval(interp,"SetupDimStackKeyDialog");

    Tcl_Eval(interp,"SetupBrushStats");

    Tcl_Eval(interp,"SetupClusterBrush");

    Tcl_Eval(interp,"SetupGlyphPlacement");

    Tcl_Eval(interp,"SetupDependentDimensionsMenu");

    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static int appSaveBrushedOkcFile(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *what = argv[1];

    dataset->WriteBrushed(what);

//    appRefreshAllCanvas(interp);
    return TCL_OK;
}

static const char *interface_item_names[] = {
                    "background",
                    "text",
                    "data",
                    "highlight",
                    "overlap",
                    "paint",
                    "grid1",
                    "grid2",
                    "grid3",
                    "stat1",
                    "stat2",
                    "brush1",
                    "brush2",
                    "brush3",
                    "brush4",
                    "sbb"
                };

static int appQueryInterfaceColor(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *what = argv[1];
    unsigned int col;

    for (int i=0;i<NUM_INTERFACE_ITEM;i++) {
        if (strcmp(what,interface_item_names[i])==0) {
            col = interface_color[i];

            char str[100];
            sprintf(str,"#%06x",col);

            Tcl_SetResult( interp, str, TCL_STATIC );
            return TCL_OK;
        }
    }

    cerr << "Wrong interface item query: " << what << "\n";
    exit(-1);
    return TCL_OK;
}

static int appSetInterfaceColor(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    const char *what = argv[1];
    const char *cstr = argv[2];

    unsigned int col;
    sscanf(cstr+1,"%x",&col);

    for (int i=0;i<NUM_INTERFACE_ITEM;i++) {
        if (strcmp(what,interface_item_names[i])==0) {

            interface_color[i] = col;

            if (dataset) BrushOperMarkDataset(); // Need to recolor the points

            return TCL_OK;
        }
    }

    cerr << "Wrong interface item query: " << what << "\n";
    exit(-1);
    return TCL_OK;
}


static int appMainInit(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    cout << "--- appMainInit\n";
    int result = Tcl_Eval(interp,"InitMainMisc");
    cout << "--- End appMainInit\n";
    assert(result == TCL_OK);

    appRegisterCanvas(interp,"structBrushCanvas","cfNav");
    appRegisterCanvas(interp,"histoCanvas","cfHist");
    appRegisterCanvas(interp,"minimapCanvas","minimap");
    appRegisterCanvas(interp,"treemapCanvas","treemap");
    appRegisterCanvas(interp,"glyphToolboxCanvas","glyphtoolbox");
    appRegisterCanvas(interp,"stackToolboxCanvas","stacktoolbox");

    return TCL_OK;
}

static int InitApplication(Tcl_Interp *interp)
{
    assert(interp);

    // Initialize TCL, TK, and TOGL.
    //
    if (Tcl_Init(interp) == TCL_ERROR) return TCL_ERROR;
    if (Tk_Init(interp) == TCL_ERROR) return TCL_ERROR;
    if (Togl_Init(interp) == TCL_ERROR) return TCL_ERROR;

    // Application UI Tcl File
    Tcl_SetVar( interp, "tcl_rcFileName", (char*) interface_file, TCL_GLOBAL_ONLY );

    // Sets up a function where we can startup our application
    //
    Togl_CreateFunc( grCanvas::grInitCanvas );
    Togl_DisplayFunc( grCanvas::grDisplayEvent );
    Togl_ReshapeFunc( grCanvas::grReshapeEvent );

    // Setup C hooks to Tk interface
    //
    #define ADD_COMMAND(X) Tcl_CreateCommand(interp, "" #X "", X, 0, 0);

    ADD_COMMAND(appMainInit);
    ADD_COMMAND(appMainCanvasRefresh);
    ADD_COMMAND(appRefreshAllCanvas);
    ADD_COMMAND(appModeSelect);
    ADD_COMMAND(appLoadOkcFile);
    ADD_COMMAND(appSaveBrushedOkcFile);
    ADD_COMMAND(appCloseOkcFile);
    ADD_COMMAND(appQueryInterfaceColor);
    ADD_COMMAND(appSetInterfaceColor);
    ADD_COMMAND(appSetCanvasZoom);
    ADD_COMMAND(appGetCanvasViewParameters);
    ADD_COMMAND(appSetViewScrollTo);
    ADD_COMMAND(appGetDatasetSummary);
    ADD_COMMAND(appQueryGlyphKey);
    ADD_COMMAND(appQueryDimStackKey);
    ADD_COMMAND(appGetFieldInfo);
    ADD_COMMAND(appGetDimensionNames);
    ADD_COMMAND(appGetPrincipalNames);
    ADD_COMMAND(appGetClusterBrush);
    ADD_COMMAND(appSetFieldOnOff);
    ADD_COMMAND(appReorderDimension);
    ADD_COMMAND(appGlyphPlacementSort);
    ADD_COMMAND(appGlyphPlacementOriginal);
    ADD_COMMAND(appGlyphPlacementDataDriven);
    ADD_COMMAND(appGlyphPlacementDerived);
    ADD_COMMAND(appDrawBrushStatsCanvas);
    ADD_COMMAND(appSetBandsOnOff);
    ADD_COMMAND(appMagnify);
    ADD_COMMAND(appDeMagnify);
    ADD_COMMAND(appSetClusterRadius);
    ADD_COMMAND(appSetBrushedRadius);
    ADD_COMMAND(appSetExtentScalingFactor);
    ADD_COMMAND(appSetDynamicMaskingScale);
    ADD_COMMAND(appGlobalBrushResize);
    ADD_COMMAND(appSetBrushEnableState);
    ADD_COMMAND(appSetBrushDisplayState);
    ADD_COMMAND(appSetBrushBoundaryType);
    ADD_COMMAND(appSetActiveBrush);
    ADD_COMMAND(appGetDimensionName);
    ADD_COMMAND(appDependentDimensionHasChanged);
    ADD_COMMAND(appSaveMainCanvasImage);
    ADD_COMMAND(appRegisterCanvas);
    ADD_COMMAND(appSetDoubleBuffer);
    ADD_COMMAND(appNewBrushOps);
    ADD_COMMAND(appDeleteBrushOps);
    ADD_COMMAND(appFormBrushExpr);
    ADD_COMMAND(appGetBrushOpString);
    ADD_COMMAND(appRaiseOperPriority);
    ADD_COMMAND(appLowerOperPriority);
    ADD_COMMAND(appGetBrushOps);
    ADD_COMMAND(appGetNumBrushes);
    ADD_COMMAND(appGetBrushExpr);
    ADD_COMMAND(appSetMaskBrushedData);
    ADD_COMMAND(appSetMaskUnbrushedData);
    ADD_COMMAND(appSetMaskBrush);
//  ADD_COMMAND(appUpdataDisplays);
    ADD_COMMAND(appSetupDataValuesDialog);

    appInitGlobals();

    return TCL_OK;
}

ofstream LOGFP("LOG");

main( int argc, char **argv )
{
LOG("XmdvTool Started in main().");

    // Search for the Tk/Tcl interface file.
    // Assumes interface file is named <executable>.tcl
    //
#ifdef WIN32
//    setenv("TCL_LIBRARY","./LIB/TCL8.0",1);
//jing
	_putenv("TCL_LIBRARY=./wintcl/lib/tcl8.1");
    strcpy(interface_file,"XmdvTool.tcl");
#else
	putenv("TCL_LIBRARY=./tcl/lib/tcl8.1");
    realpath(argv[0],interface_file);
    strcat(interface_file,".tcl");
#endif
    struct stat buf;
    int exists = stat(interface_file,&buf);
    if (exists!=0) {
        cerr << "Missing Interface Definition file: \""<<interface_file<<"\".\n";
        exit(-1);
    }

    Tk_Main( argc, argv, InitApplication);
    assert(0); // We should never get here.
}
