////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef BRUSH_H
#define BRUSH_H

#include "ARR.H"

class Brush
{
    public:
        enum BoundaryType { STEP, RAMP };

        Brush() :
            enable(false),
            display(true),
            bound_type(STEP)
        {}

        bool IsPointCovered(const ARR<float> &point);
        void Resize(float factor);

        float Min(int d);
        float Max(int d);
        float RampMin(int d);
        float RampMax(int d);
        void SetMinMax(int d,float lo,float hi);

    public:
        int num;
        ARR<float> pos;
        ARR<float> size;
        bool enable;              // if this brush is on
        bool display;             // display this brush?
        BoundaryType bound_type;       // brush boundary type
        ARR<float> bound_offset;  // boundary offset from min/max
};

const int MAXBRUSH = 4;
extern Brush brushes[MAXBRUSH], *curr_brush, all_brushes[MAXBRUSH];
extern unsigned long brush_color[MAXBRUSH];

extern void InitBrushSet();
extern int SetActiveBrush(int which);
extern int SetBrushEnableState(int which, bool state);
extern int SetBrushDisplayState(int which, bool state);
extern int SetBrushBoundaryType(int which, Brush::BoundaryType type);
extern bool CheckBrushCoverage(const ARR<float> &point, Brush *brush);
extern float PercentBrushCoverage(const ARR<float> &point, Brush *brush);
extern int CoverageRank(const ARR<float> &point);

#endif
