////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef CLUSTER_H
#define CLUSTER_H

#include "ARR.H"
#include "okcdata.H"

#define IS_CONTOUR0  1<<0
#define IS_CONTOUR1  1<<1
#define IS_BRUSHED   1<<2
#define IS_MAGNIFIED 1<<3

struct Cluster 
{
    int parent;
    int num_children;
    ARR<int> child;

    ARR<float> mean;
//jing
    ARR<float> max;
    ARR<float> min;
//jing
    ARR<float> all_mean;

    okcdata::mark_type mark;

    // cfentry of a cluster
    int entries;
    ARR<float> sx;
    float sxx;
    float sqrt_radius;
    float sqrt_diameter;

    float color;
    
    // The dimensions for the rect in tree-map
    float xmin;
    float xmax;
    float ymin;
    float ymax;

    int flags;              
};

struct ClusterTree
{
    char *clusterfile;
    
    ARR<Cluster> nodes;
    int dimension;
    
    ARR<float> min_extents;
    ARR<float> max_extents;

    ClusterTree() :
        clusterfile(0),
        dimension(0)
    {}

    ~ClusterTree() {
        if (clusterfile) delete [] clusterfile;
    }   

    bool ReadClusterTreeFile(const char* fname);
    bool ReadClusterTreeFile2();
    bool ReadClusterTreeFile3();
};

extern ClusterTree *btree;

extern void hierDeleteClusterTree();
extern void hierInitExtents();
extern void hierInitClusterTree();
extern void hierClearNodeFlags(int bits=0xffffff);
extern void hierMarkContourNodes(float radius, int FLAG);
extern void hierModifyExtentsByBrushParameters();
extern void hierModifyBrushParametersByExtents();
extern void hierMarkBrushedClustersByColor(Cluster* node, float L, float R);
extern void hierMarkBrushedClustersByExtents(Cluster* node);
extern void hierSetMeanOnOff();
extern void hierReorderDimension();

#endif
