////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_dimstack.H"
#include "cluster_parcoord.H"
#include "ogl.H"
#include "cluster.H"
#include "XmdvTool.H"
#include "colormap.H"
#include "dimstack.H"
#include "magnify.H"
#include "brush_oper.H"

//variables for drawing
static ARR<float> size;
static ARR<float> data;
static ARR<float> mean;

//static ARR<float> max_mean;
//static ARR<float> min_mean;
static ARR<float> max;
static ARR<float> min;

static float x_blocksize, y_blocksize, x_size, y_size;

static ARR<int> contrib0;
static ARR<int> contrib1;
static ARR<int> contrib2;
static ARR<int> contrib;

static grCanvas *canvas;
static RGBt color;
static float alpha;


void RegisterHierDimstackCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_cluster_dimstack;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputHierDimstackMouseMotion;
    canvas->buttonpress_proc = InputHierDimstackButtonPress;
    canvas->buttonrelease_proc = InputHierDimstackButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = InputHierDimstackKeyRelease;
}

/*
static float distanceFactor(ARR<float> data)
{
	float temp, temp1, factor, sum = 0;
	for ( int i = 0; i < dataset->dims; i++ )
	{
		temp = data[i] - mean[i];
		if ( temp > 0 )
			temp1 = max_mean[i]; 
		else
			temp1 = min_mean[i]; 
		if ( temp1 > 1e-6 )
			factor = temp / temp1;
		else
			factor = 1;
		sum += factor;
	}
	factor /= dataset->dims;
	if ( factor < 0.0 )
		factor = -1.0 * factor;
	return factor*alpha;
}
*/

static bool is_within(int val, int dim, int brush)
{
    float fval = val;
    float range = dataset->dim_max[dim] - dataset->dim_min[dim];
    float b_pos = dataset->cardinality[dim] * (brushes[brush].pos[dim]-dataset->dim_min[dim]) / range;
    float b_size = .5 * dataset->cardinality[dim] * brushes[brush].size[dim] / range;
    float start = b_pos - b_size;
    float end = b_pos + b_size;
    if (end > fval && end <= (fval+1.0)) return true;
    if (start >= fval && start < (fval+1.0)) return true;
    if (start < fval && end > (fval+1.0)) return true;
    return false;
}

static void rec_fill(grCanvas *canvas, float basex, float basey, int nextx, int nexty,
          ARR<float> &size, int brush)
{
    // If we had an even number of dimensions and are at the bottom of the
    // recursion, draw a rectangle in that data spot, as it is covered.
    //
    if (nextx == dataset->dims)
    {
        canvas->FillRectangle(basex, y_size - basey - y_blocksize, x_blocksize, y_blocksize);
        return;
    }

    // If we had an odd number of dimensions and are at the bottom of the
    // recursion, check each descrete value for brush overlap and draw rectangles
    // over covered spots.
    //
    int i;
    if (nexty == dataset->dims)
    {
        for (i=0; i<dataset->cardinality[nextx]; i++) {
            if (is_within(i, nextx, brush) == true)
                canvas->FillRectangle(basex + i * size[nextx],
                    y_size - basey - y_blocksize, x_blocksize, y_blocksize);
        }
        return;
    }

    // Continue recursion for each bucket of the current 2 dimensions
    //
    for (i=0; i<dataset->cardinality[nextx]; i++)
        for (int j=0; j<dataset->cardinality[nexty]; j++)
            if ((is_within(i, nextx, brush) == 1) && (is_within(j, nexty, brush) == 1))
                rec_fill(canvas, basex + i * size[nextx], basey + j * size[nexty],
                    nextx+2, nexty+2, size, brush);
}


static void comp_pixel_size()
{
    float odd=1, even=1;

    // Compute product of cardinalities in each orientation
    int i;
    for (i = 0;i < dataset->dims;i=i+2)
        odd *= dataset->cardinality[i];

    for (i = 1;i < dataset->dims;i=i+2)
        even *= dataset->cardinality[i];

    odd = 1.0/odd;
    even = 1.0/even;

    // Now compute size of a virtual pixel for each dimension
    for (i = 0;i < dataset->dims;i++)   {
        size[i] = (i&1)?even:odd;
        for (int j = i+2; j < dataset->dims; j=j+2)
            size[i] *= dataset->cardinality[j];
    }

    x_blocksize = odd;
    y_blocksize = even;
}

static void ds_outline(grCanvas *canvas, ARR<float> &size)
{
    for (int brush=0; brush<MAXBRUSH; brush++)
    {
        // If this brush is not being displayed, skip it
        if (brushes[brush].display == false) continue;

        // Set the appropriate brush color
        canvas->SetForeground(interface_color[BRUSH1+brush]);

        // For the outermost dimensions...
        for (int i=0; i<dataset->cardinality[0]; i++) {
            for (int j=0; j<dataset->cardinality[1]; j++)
            {
                // if the brush intersects those dimensions recursively
                // call to fill
                if ((is_within(i, 0, brush) == 1) && (is_within(j, 1, brush) == 1))
                    rec_fill(canvas, i * size[0], j * size[1], 2, 3, size, brush);
            }
        }
    }
}

//
// Compute the descretized value from an original data value for a particular
// dimension.  Display will be messed up if it is out of bounds.  You then
// should fix the range for that dimension in your data file.
//
static int get_contrib(int ind, float data)
{
    int res = int( ((data - dataset->dim_min[ind]) / (dataset->dim_max[ind] - dataset->dim_min[ind])) *
              (float)(dataset->cardinality[ind]));

    if (res >= dataset->cardinality[ind] )
	{
		res--;
	}
	else if ( res < 0) 
	{
		res++;
	}
/*
    if (res >= dataset->cardinality[ind] || res < 0) {
        printf("Data point out of bounds: dimension %d: [%f, %f]: %f\n", ind,
            dataset->dim_min[ind], dataset->dim_max[ind], data);
    }
*/
    return res;
}

static void DrawSingleDimstackBin(grCanvas *canvas, const ARR<float> &data, const ARR<float> &size)
{
    ARR<int> contrib(dataset->dims);
    int j;
    for (j=0; j<dataset->dims; j++)
        contrib[j] = get_contrib(j, data[j]);

    float endx = 0, endy = 0;
    for (j=0; j<dataset->dims; j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < dataset->dims)
            endy += contrib[j] * size[j];
    }
    canvas->FillRectangle(endx, y_size - endy - y_blocksize, x_blocksize, y_blocksize);
}

//calculate the distance factor of a bin in the band
static float distanceFactor()
{
	float temp, temp1;
	float factor, sum = 0;
	for ( int i = 0; i < dataset->dims; i++ )
	{
		if ( contrib[i] > contrib0[i] )
		{
			temp = contrib[i] - contrib0[i];
			temp1 = contrib2[i] - contrib0[i];
		}
		else
		{
			temp = contrib0[i] - contrib[i];
			temp1 = contrib0[i] - contrib1[i];
		}
		if ( temp1 > 0 )
			factor = temp / temp1;
		else
			factor = 0;
		sum += factor;
	}
	factor = sum / dataset->dims;
	if ( factor < 0.0 )
		factor = -1.0 * factor;

	if ( fabs( factor ) < 1e-6 )
		return 0; 
	else
		return ((1-factor)*0.9+0.1)*alpha;
}

//draw a bin in a band
static void DrawSingleDimstackBin2()
{
	double density = distanceFactor();
	canvas->SetForeground( color, density );

    float endx = 0, endy = 0;
    for (int j=0; j<dataset->dims; j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < dataset->dims)
            endy += contrib[j] * size[j];
    }
    canvas->FillRectangle(endx, y_size - endy - y_blocksize, x_blocksize, y_blocksize);
}

//draw band of level k
static void traceBand(int k)
{
	int j;
	if ( k == dataset->dims-1 )
	{
		for ( j = contrib1[k]; j <= contrib2[k]; j++ )
		{
			contrib[k] = j;
			DrawSingleDimstackBin2();
		}
		return;
	}
	else
	{
		for ( j = contrib1[k]; j <= contrib2[k]; j++ )
		{
			contrib[k] = j;
			traceBand(k+1);
		}
		return;
	}
}

//draw a complete band
static void hierDimstackDrawSingleNodeBand2(const ARR<float> &data0, const ARR<float> &data1, const ARR<float> &data2)
{
    int j;
    for (j=0; j<dataset->dims; j++)
        contrib0[j] = get_contrib(j, data0[j]);
    for (j=0; j<dataset->dims; j++)
        contrib1[j] = get_contrib(j, data1[j]);
    for (j=0; j<dataset->dims; j++)
        contrib2[j] = get_contrib(j, data2[j]);

	traceBand(0);
}

/*
static void hierDimstackDrawSingleNodeBand2(Cluster &c)
{
	if ( c.num_children == 0 )
	{
		float density; 
		density = distanceFactor(c.mean);
		canvas->SetForeground( color, density );
        DrawSingleDimstackBin(canvas,c.mean,size);
		return;
	}
	else
	{
		for ( int i = 0; i < c.num_children; i++ )
		{
	        Cluster &c1 = btree->nodes[c.child[i]];
			hierDimstackDrawSingleNodeBand2(c1);
		}
	}
}*/

//draw the band of a cluster
static void hierDimstackDrawSingleNodeBand( grCanvas *canvas1, Cluster &c, float density)
{
	canvas = canvas1;
	bool is_brushed = BITTEST(c.flags,IS_BRUSHED);
    if (MAGNIFY && !is_brushed) return;
    if ( c.num_children == 0 ) return;
	//set color
    if (COLORED_BANDS) {

        int color_idx = (int)(c.color*(CMAPSIZE-1));

        if (is_brushed)  {
            DB(fade_factor);
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1+fade_factor);
        }
        else {
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1-fade_factor);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
        }
    }
    else
        canvas->SetForeground(interface_color[BRUSH1+2]);

	//draw the band for the cluster
	for ( int i = 0; i < dataset->dims; i++ )
	{
		mean[i] = c.mean[i];
		max[i] = (c.max[i]-c.mean[i])*extent_scaling_fac + c.mean[i];
		min[i] = (c.min[i]-c.mean[i])*extent_scaling_fac + c.mean[i];
	}
	
	color = canvas->fg_color;
	alpha = canvas->alpha;
//	hierDimstackDrawSingleNodeBand2(c);
	hierDimstackDrawSingleNodeBand2(mean, min, max);

}


// jing draw band of all clusters
static void hierDimstackDrawColoredBands(grCanvas *canvas, int on_flags=0, int off_flags=0)
{
	for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierDimstackDrawSingleNodeBand(canvas, c, 0.5);
    }
}

//draw the mean for one cluster
static void hierDimstackDrawSingleNodeMean( grCanvas *canvas, Cluster &c, int style)
{
    if (BITTEST(style,STYLE_NONE)) return;

    //set color
	float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }

	//draw the mean
	for ( int i = 0; i < dataset->dims; i++ )
	{
		data[i] = c.mean[i];
	}

    DrawSingleDimstackBin(canvas,data,size);

}

//draw means for all clusters
static void hierDimstackDrawColoredMeans(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;
	    hierDimstackDrawSingleNodeMean(canvas, c, style);
    }
}

void do_cluster_dimstack( grCanvas *canvas )
{
    if (!dataset) return;

    if (dataset->dims<2) {
        fprintf(stderr,"Warning: Dimension stacking needs at least 2 active dimensions.\n");
        return;
    }

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    hierInitClusterTree();

    if (!btree) return;

    size.SetLen(dataset->dims);
    data.SetLen(dataset->dims);
    mean.SetLen(dataset->dims);
    max.SetLen(dataset->dims);
    min.SetLen(dataset->dims);
//    max_mean.SetLen(dataset->dims);
//    min_mean.SetLen(dataset->dims);

    contrib0.SetLen(dataset->dims);
    contrib1.SetLen(dataset->dims);
    contrib2.SetLen(dataset->dims);
    contrib.SetLen(dataset->dims);

    comp_pixel_size();
    x_size = size[0]*dataset->cardinality[0];
    y_size = size[1]*dataset->cardinality[1];

    // Display the brush coverage if desired
//    ds_outline(canvas, size);

	//draw grid 
    // Draw up to 3 levels of grid marks, based on the number of dimensions
    if (dataset->dims >= 6)
    {
        canvas->SetForeground(interface_color[GRID3]);
        for (float x = 0;x <= x_size; x = x + size[4])
            canvas->DrawLine(x, 0, x, y_size);

        for (float y = 0;y <= y_size; y = y + size[5])
            canvas->DrawLine(0, y, x_size, y);
    }

    if (dataset->dims >= 4)
    {
        canvas->SetForeground(interface_color[GRID2]);
        for (float x = 0;x <= x_size; x = x + size[2])
            canvas->DrawLine(x, 0, x, y_size);

        for (float y = 0;y <= y_size; y = y + size[3])
            canvas->DrawLine(0, y, x_size, y);
    }

    canvas->SetForeground(interface_color[GRID1]);

    for (float x = 0;x <= x_size; x = x + size[0])
        canvas->DrawLine(x, 0, x, y_size);

    for (float y = 0;y <= y_size; y = y + size[1])
        canvas->DrawLine(0, y, x_size, y);


    if (SHOW_MEANS) {
        if (!MAGNIFY) hierDimstackDrawColoredMeans(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
        hierDimstackDrawColoredMeans(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
    }
	

    if (SHOW_BANDS) {
        if (!MAGNIFY) hierDimstackDrawColoredBands(canvas, IS_CONTOUR0);
        hierDimstackDrawColoredBands(canvas, IS_CONTOUR1 | IS_BRUSHED);
    }

}


void InputHierDimstackMouseMotion(grCanvas *canvas, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && canvas->button[0] == true) return;

    InputDimstackMouseMotion(canvas,x,y);
}

void InputHierDimstackButtonPress(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputDimstackButtonPress(canvas,button,x,y);
}

void InputHierDimstackButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputDimstackButtonRelease(canvas,button,x,y);

    if (canvas->shiftkey) return;

    if (!btree) return;

//    EXTENT_BRUSH = true;
//    hierClearNodeFlags(IS_BRUSHED);
//    hierModifyExtentsByBrushParameters();
//    hierMarkBrushedClustersByExtents(&btree->nodes[0]);
    canvas->Refresh();
}

#include "X11/keysym.h"

void InputHierDimstackKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) return;

    appCanvasInputKeyRelease(canvas,key);

    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        //
//        EXTENT_BRUSH = true;
//        hierClearNodeFlags(IS_BRUSHED);
//        hierModifyExtentsByBrushParameters();
//        hierMarkBrushedClustersByExtents(&btree->nodes[0]);
        canvas->Refresh();
    }
}
