////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_glyph.H"
#include "cluster_parcoord.H"
#include "ogl.H"
#include "cluster.H"
#include "XmdvTool.H"
#include "colormap.H"
#include "glyph.H"
#include "magnify.H"
#include "brush_oper.H"

//the varialbes for drawing
static float blocksize;
static int blockcnt;  
static float halfblocksize;

static float basex;
static float basey;
static double *isin;
static double *icos;
static double alpha;
static ARR<Vec2> polymean;
static ARR<Vec2> polymax;
static ARR<Vec2> polymin;
static int count;
 
void RegisterHierGlyphCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_cluster_glyph;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputHierGlyphMouseMotion;
    canvas->buttonpress_proc = InputHierGlyphButtonPress;
    canvas->buttonrelease_proc = InputHierGlyphButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = InputHierGlyphKeyRelease;
}

//draw the band of a cluster
static void hierGlyphDrawSingleNodeBand( grCanvas *canvas, Cluster &c, float density)
{
	bool is_brushed = BITTEST(c.flags,IS_BRUSHED);
    if (MAGNIFY && !is_brushed) return;
    if ( c.num_children == 0 ) return;
	//Set color
    if (COLORED_BANDS) {

        int color_idx = (int)(c.color*(CMAPSIZE-1));

        if (is_brushed)  {
            DB(fade_factor);
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1+fade_factor);
        }
        else {
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1-fade_factor);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
        }
    }
    else
        canvas->SetForeground(interface_color[BRUSH1+2]);

	double temp1, temp2, temp3;
	int i;

	polymean[dimension] = polymean[0];
	
	for ( i = 0; i < dimension; i++ )
	{
        temp1 = (c.mean[i] - dataset->dim_min[i]);
        temp2 = (c.max[i] - dataset->dim_min[i]);
        temp3 = (c.min[i] - dataset->dim_min[i]);
		if ( fabs(1 - extent_scaling_fac ) > 1e-4 )
		{
			temp2 = temp1 + extent_scaling_fac*(temp2-temp1);
			temp3 = temp1 + extent_scaling_fac*(temp3-temp1);
        }
		polymax[i].X = basex + temp2 * icos[i];
        polymax[i].Y = basey - temp2 * isin[i];;
		polymin[i].X = basex + temp3 * icos[i];
        polymin[i].Y = basey - temp3 * isin[i];;
	}
	polymax[dimension] = polymax[0];
	polymin[dimension] = polymin[0];

    ARR<Vec2> band(4);

    // Draw the lower band
    int d;
    for (d=0; d<dimension+1; d++)
    {
		Vec2 lo, hi;

        lo = polymin[d];

		hi = polymean[d];

//        lo.Y = (1-extent_scaling_fac)*hi.Y + (extent_scaling_fac*lo.Y);

        if (d==0) {
            band[0] = lo;
            band[3] = hi;

        } else {
	
            band[1] = lo;
            band[2] = hi;
            canvas->DrawTransparentBand(band, 0.0, density, canvas->fg_color, canvas->fg_color);

            band[0] = band[1];
            band[3] = band[2];
        }
    }

    // Draw the upper band
    for (d=0; d<dimension+1; d++)
    {
        Vec2 lo, hi;

		lo = polymean[d];
		hi = polymax[d];

        // Calculate how much to scale, if any
//        hi.Y = (1-extent_scaling_fac)*lo.Y + (extent_scaling_fac*hi.Y);

        if (d==0) {
            band[0] = lo;
            band[3] = hi;
        } else {
            band[1] = lo;
            band[2] = hi;

            canvas->DrawTransparentBand(band, density, 0.0, canvas->fg_color, canvas->fg_color);

            band[0] = band[1];
            band[3] = band[2];
        }
    }
}

/*
static void hierGlyphDrawColoredBands(grCanvas *canvas, int on_flags=0, int off_flags=0)
{
//jing draw cluster's band one by one
	for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;
        basex = blocksize * (count%blockcnt) + halfblocksize;
        basey = blocksize * (count/blockcnt) + halfblocksize;
		count ++;
        hierGlyphDrawSingleNodeBand(canvas, c, 0.5);
    }
}
*/

//draw mean and band for a cluster
static void hierGlyphDrawSingleNodeMean( grCanvas *canvas, Cluster &c, int style)
{
	int i;
	// jing setting
    if (BITTEST(style,STYLE_NONE)) return;

    float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }

	//draw the mean of the cluster
	double temp;
	for ( i = 0; i < dimension; i++ )
	{
        temp = (c.mean[i] - dataset->dim_min[i]);
		polymean[i].X = basex + temp * icos[i];
        polymean[i].Y = basey - temp * isin[i];;
	}
	canvas->BeginDraw();

	glBegin(GL_LINES);
    for (i =0; i < dimension; i++) 
	{
		glVertex2f(basex, basey);
		glVertex2f(polymean[i].X, polymean[i].Y);
    }
	glEnd();

	glBegin(GL_LINE_LOOP);
    for (i =0; i < dimension; i++) 
	{
		glVertex2f(polymean[i].X, polymean[i].Y);
    }
	glEnd();

	canvas->EndDraw();
	
	//draw the band of a cluster
	if ( SHOW_BANDS )
		hierGlyphDrawSingleNodeBand(canvas, c, 0.5);

}

//draw means and bands for all clusters
static void hierGlyphDrawColoredMeans(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
	for (int i=0;i< btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;
        basex = blocksize * (count%blockcnt) + halfblocksize;
        basey = blocksize * (count/blockcnt) + halfblocksize;
		count ++;
	    hierGlyphDrawSingleNodeMean(canvas, c, style);
    }
}

//prepare for drawing
void hierGlyphSetParameter(int on_flags1=0, int off_flags1=0, int on_flags2=0, int off_flags2=0)
{
	int nodeNumber = 0, i; 

	if (!MAGNIFY)
	{
		for (i=0;i<btree->nodes.Len();i++)
		{
			Cluster &c = btree->nodes[i];
			if (!SCAN_FLAGS(c.flags, on_flags1, off_flags1)) continue;
			nodeNumber ++;
		}
	}
    
	for (i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags2, off_flags2)) continue;
		nodeNumber ++;
    }

    
	dimension = dataset->dims;
	//set canvas arrangement parameters
	blocksize = 1.0/(1.0 + sqrt(nodeNumber));
    blockcnt = int( .5 + sqrt(nodeNumber) );
    halfblocksize = .5 * blocksize;

    alpha = 2.0*M_PI/dimension;
	isin = new double[dimension];
	icos = new double[dimension];
    
	double temp;
	for (i=0; i<dimension; i++)
    {
        temp = (blocksize/2.0) / (dataset->dim_max[i] - dataset->dim_min[i]);
		icos[i] = temp * cos(i*alpha);
        isin[i] = temp * sin(i*alpha);
	}

	polymean.SetLen(dimension+1);
	polymax.SetLen(dimension+1);
	polymin.SetLen(dimension+1);
}

//free space after drawing
void hierGlyphFreeSpace()
{
	delete icos;
	delete isin;
}


void do_cluster_glyph( grCanvas *canvas )
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    hierInitClusterTree();

    if (!btree) return;

	
	//draw brushes    
	//if (SHOW_BRUSH_BKGRD)
//?    DrawScatterBrushes(canvas);

	hierGlyphSetParameter(IS_CONTOUR0, IS_BRUSHED, IS_CONTOUR1 | IS_BRUSHED);

    if (SHOW_MEANS) {
		glLineWidth(DSIZE);
		count = 0;
        hierGlyphDrawColoredMeans(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
        if (!MAGNIFY) hierGlyphDrawColoredMeans(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
	    glLineWidth(1.0);
    }
/*
    if (SHOW_BANDS) {
		count = 0;
        hierGlyphDrawColoredBands(canvas, IS_CONTOUR1);
        if (!MAGNIFY) hierGlyphDrawColoredBands(canvas, IS_CONTOUR0);
    }
*/	
	hierGlyphFreeSpace();

	/*

*/
 
	
	/*
    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set 
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dimension);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawParcoordAverage(canvas,average);
    }

    DrawParcoordAxes(canvas);*/
}


void InputHierGlyphMouseMotion(grCanvas *canvas, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && canvas->button[0] == true) return;

    InputGlyphMouseMotion(canvas,x,y);
}

void InputHierGlyphButtonPress(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputGlyphButtonPress(canvas,button,x,y);
}

void InputHierGlyphButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputGlyphButtonRelease(canvas,button,x,y);

    if (canvas->shiftkey) return;

    if (!btree) return;

//    EXTENT_BRUSH = true;
//    hierClearNodeFlags(IS_BRUSHED);
//    hierModifyExtentsByBrushParameters();
//    hierMarkBrushedClustersByExtents(&btree->nodes[0]);
    canvas->Refresh();
}

#include "X11/keysym.h"

void InputHierGlyphKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) return;

    appCanvasInputKeyRelease(canvas,key);

    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        //
//        EXTENT_BRUSH = true;
//        hierClearNodeFlags(IS_BRUSHED);
//        hierModifyExtentsByBrushParameters();
//        hierMarkBrushedClustersByExtents(&btree->nodes[0]);
        canvas->Refresh();
    }
}
