////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the		
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_parcoord.H"
#include "ogl.H"
#include "cluster.H"
#include "XmdvTool.H"
#include "colormap.H"
#include "parcoord.H"
#include "magnify.H"
#include "brush_oper.H"


//the varialbes for drawing a polygonal line
ARR<Vec2> poly;
double* mapedX;
double* mapedYMean;
double* mapedYMax;
double* mapedYMin;
double* para;
Vec2 lo, hi;

//the variables controlling Cluster Hierarchical Toolbox 
float cluster_radius;
colormap CMAP(true);  
bool EXTENT_BRUSH;
float fade_factor;
int normal_style;
int highlight_style;
bool SHOW_BANDS;
bool SHOW_MEANS;
bool SHOW_BRUSH_BKGRD;
bool COLORED_BANDS;
float extent_scaling_fac;
enum COLOR_STYLE;
float tree_brush[2];
float brushed_radius;
int tree_height;

int dimension;


//Initialize the variables controlling Cluster Hierarchical Toolbox. 
//Used when open a new file
void InitClusterSet()
{
	cluster_radius = 0;
	brushed_radius= 0;
	fade_factor=0; // [-1,+1]
	extent_scaling_fac=1.0;

	SHOW_BANDS = false;//shou_bands

	SHOW_MEANS = true;
	SHOW_BRUSH_BKGRD = true;
	COLORED_BANDS = true;

	normal_style = STYLE_COLORMAP;
	highlight_style = STYLE_HIGHLIGHT | STYLE_THICK;

	tree_brush[0] = .25;
	tree_brush[1] = .75;
	tree_height=0;
}

//Used when map data to the canvas.
#define XMAP(I) (((I)+.5) * dataset->one_over_dims)
//#define YMAP(X,Y) (1.0-((Y) - dataset->dim_min[X]) * para[X])
#define YMAP(X,Y) (1.0-((Y) - dataset->dim_min[X]) / (dataset->dim_max[X] - dataset->dim_min[X]))

#define YINVMAP(X,Y) (((1.0-(Y)) * (dataset->dim_max[X] - dataset->dim_min[X])) + dataset->dim_min[X])
#define YGAP(X,Y) ((Y)/(dataset->dim_max[X] - dataset->dim_min[X]))

void RegisterHierParcoordCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_cluster_parcoord;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputHierParcoordMouseMotion;
    canvas->buttonpress_proc = InputHierParcoordButtonPress;
    canvas->buttonrelease_proc = InputHierParcoordButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = InputHierParcoordKeyRelease;
}

//draw a band for one cluster
static void hierDrawSingleNodeBand( grCanvas *canvas, Cluster &c, float density)
{
	bool is_brushed = BITTEST(c.flags,IS_BRUSHED);
    if (MAGNIFY && !is_brushed) return;

	//Set color
    if (COLORED_BANDS) {

        int color_idx = (int)(c.color*(CMAPSIZE-1));

        if (is_brushed)  {
            DB(fade_factor);
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1+fade_factor);
        }
        else {
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1-fade_factor);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
        }
    }
    else
        canvas->SetForeground(interface_color[BRUSH1+2]);

    // Draw the lower band
    int d;

    lo.X = mapedX[0];
    lo.Y = mapedYMin[0];

    hi.X = lo.X;
    hi.Y = mapedYMean[0];

	ARR<Vec2> band(4);

    band[0] = lo;
    band[3] = hi;
    for (d=1; d<dimension; d++)
    {
        lo.X = mapedX[d];
        lo.Y = mapedYMin[d];

        hi.X = lo.X;
        hi.Y = mapedYMean[d];

        band[1] = lo;
        band[2] = hi;
        canvas->DrawTransparentBand(band, 0.0, density, canvas->fg_color, canvas->fg_color);

        band[0] = band[1];
        band[3] = band[2];
    }

    // Draw the upper band
    lo.X = mapedX[0];
    lo.Y = mapedYMean[0];

    hi.X = lo.X;
    hi.Y = mapedYMax[0];
    band[0] = lo;
    band[3] = hi;

    for (d=1; d<dimension; d++)
    {
        lo.X = mapedX[d];
        lo.Y = mapedYMean[d];

        hi.X = lo.X;
        hi.Y = mapedYMax[d];

		band[1] = lo;
        band[2] = hi;

        canvas->DrawTransparentBand(band, density, 0.0, canvas->fg_color, canvas->fg_color);

        band[0] = band[1];
        band[3] = band[2];
	}
}

//draw the mean of one cluster
static void hierDrawSingleNodeMean( grCanvas *canvas, Cluster &c, int style)
{
    //set color
	if (BITTEST(style,STYLE_NONE)) return;

    float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }

	//map the mean of a cluster to a polygonal line
	double temp = dataset->dim_max[3];

	int d;
    for (d=0; d<dimension; d++)
    {
        poly[d].X = mapedX[d];
        poly[d].Y = mapedYMean[d] = YMAP(d,c.mean[d]);
    }

	//draw the polygonal line
    canvas->DrawLines(poly);

	if ( SHOW_BANDS )
	{
		for (d=0; d<dimension; d++)
		{
			mapedYMax[d] = YMAP(d,c.max[d]);
			mapedYMin[d] = YMAP(d,c.min[d]);
			if ( fabs(1 - extent_scaling_fac ) > 1e-4 )
			{
				mapedYMax[d] = mapedYMean[d] + extent_scaling_fac * (mapedYMax[d] - mapedYMean[d]);
				mapedYMin[d] = mapedYMean[d] + extent_scaling_fac * (mapedYMin[d] - mapedYMean[d]);
			}
		}

		hierDrawSingleNodeBand(canvas, c, 0.5);
	}
}



//draw bands for all clusters
static void hierDrawColoredBands(grCanvas *canvas, int on_flags=0, int off_flags=0)
{
	for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierDrawSingleNodeBand(canvas, c, 0.5);
    }
}

//draw means for all clusters
static void hierDrawColoredMeans(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
    //set line width
	if (BITTEST(style,STYLE_THICK)) {
        glLineWidth(1.5);
    } else {
        glLineWidth(1.0);
    }

    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierDrawSingleNodeMean(canvas, c, style);
    }
    
	//reset line width
	glLineWidth(1.0);
}


static void hierParcoordSetParameter()
{
	dimension = dataset->dims;
	poly.SetLen(dimension);
	mapedX = new double [dimension];
	mapedYMean = new double [dimension];
	mapedYMax = new double [dimension];
	mapedYMin = new double [dimension];
	para = new double [dimension];
	for ( int d = 0; d < dimension; d++ )
	{
		mapedX[d] = XMAP(d);
		para[d] = 1./(dataset->dim_max[d] - dataset->dim_min[d]);
	}
}


static void hierParcoordFreeSpace()
{
	delete mapedX;
	delete mapedYMean;
	delete mapedYMax;
	delete mapedYMin;
}

//jing display_proc of hierparcoordcanvas
void do_cluster_parcoord( grCanvas *canvas )
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    hierInitClusterTree();

    if (!btree) return;

	hierParcoordSetParameter();

	//jing draw brushes    
	//if (SHOW_BRUSH_BKGRD)
//    DrawParcoordBrushes(canvas);

//jing draw bands    
/*	if (SHOW_BANDS) {
        if (!MAGNIFY) hierDrawColoredBands(canvas, IS_CONTOUR0);
        hierDrawColoredBands(canvas, IS_CONTOUR1 | IS_BRUSHED);
    }
*/

//jing draw means
    if (SHOW_MEANS) {
        if (!MAGNIFY) hierDrawColoredMeans(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
        hierDrawColoredMeans(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
    }

	hierParcoordFreeSpace();

	/*
//jing useless temporarily
    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set 
        if (!(brush_ops[op].oper & AVERAGE)) 
			continue;

        ARR<float> average(dimension);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawParcoordAverage(canvas,average);
    }
*/

    glLineWidth(1.);

    DrawParcoordAxes(canvas);
}

void InputHierParcoordMouseMotion(grCanvas *canvas, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && canvas->button[0] == true) return;

    InputParcoordMouseMotion(canvas,x,y);
}

void InputHierParcoordButtonPress(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputParcoordButtonPress(canvas,button,x,y);
}

void InputHierParcoordButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputParcoordButtonRelease(canvas,button,x,y);

    if (canvas->shiftkey) return;

    if (!btree) return;

//    EXTENT_BRUSH = true;
//    hierClearNodeFlags(IS_BRUSHED);
//    hierModifyExtentsByBrushParameters();
//    hierMarkBrushedClustersByExtents(&btree->nodes[0]);
    canvas->Refresh();
}

//jing this func need to be added
#include "X11/keysym.h"

void InputHierParcoordKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) return;

    appCanvasInputKeyRelease(canvas,key);

    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        //
//        EXTENT_BRUSH = true;
//        hierClearNodeFlags(IS_BRUSHED);
//        hierModifyExtentsByBrushParameters();
//        hierMarkBrushedClustersByExtents(&btree->nodes[0]);
        canvas->Refresh();
    }
}
