////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "colormap.H"
#include "ogl.H"

colormap::colormap(bool color=true)
{
    SetLen(CMAPSIZE);

    if (color) InstallPredefined2();
    else InstallPredefinedGreyScale();
}

void colormap::InstallPredefined1()
{
    const int NUM_RAMPS=7;
    const int RAMP[NUM_RAMPS][4] = {
        { 0,     200, 255, 255 },
        { 42,    0, 0, 255 },
        { 84,    0, 255, 255 },
        { 126,   0, 255, 0 },
        { 168,   255, 255, 0 },
        { 210,   255, 0, 255 },
        { 255,   50, 200, 200 }
    };
    InstallColorRamp(RAMP,NUM_RAMPS);
}

void colormap::InstallPredefined2()
{
    const int NUM_RAMPS=7;
    const int RAMP[NUM_RAMPS][4] = {
        { 0,     200, 255, 255 },
        { 42,    0, 0, 255 },
        { 84,    0, 255, 255 },
        { 126,   0, 255, 0 },
        { 168,   200, 255, 0 },
        { 210,   200, 0, 255 },
        { 255,   100, 200, 200 }
    };

    InstallColorRamp(RAMP,NUM_RAMPS);
}

void colormap::InstallPredefinedGreyScale()
{
    const int NUM_RAMPS=2;
    const int RAMP[NUM_RAMPS][4] = {
        { 0,      255, 255, 255 },
        { 255,    0, 0, 0 }
    };
    InstallColorRamp(RAMP,NUM_RAMPS);
}

void colormap::InstallColorRamp(const int ramp[][4], int N)
{
    for (int i=0;i<N-1;i++) 
    {
        int lo = ramp[i][0];
        int hi = ramp[i+1][0];
        
        for (int j=lo;j<=hi;j++) 
        {
            double interp = (j-lo)/(double)(hi-lo);
            int C[3];
            for (int k=0;k<3;k++) {
                C[k] = (int)((1-interp)*ramp[i][k+1] + (interp)*ramp[i+1][k+1]);
            }
            (*this)[j] = C[2]|(C[1]<<8)|(C[0]<<16);
        }
    }
}

void colormap::DrawColorBar()
{
    glShadeModel(GL_SMOOTH);
    glLineWidth(5.0);
    glBegin(GL_LINE_STRIP);
    float step = 1/float(CMAPSIZE-1);
    for (int i=0;i<CMAPSIZE;i++) 
    {
        unsigned long color = (*this)[i];
        glColor3ub((color>>16)&0xff,(color>>8)&0xff,color&0xff);
        float k = 2*i*step - 1;
        glVertex2f(k,0);
    }
    glEnd();
    glLineWidth(1.0);
}

