////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "magnify.H"
#include "XmdvTool.H"
#include "brush.H"
#include "ogl.H"
#include "brush_oper.H"

int MAGNIFY=0;

static int AnimateMagnify(grCanvas *canvas,
    const ARR<float> &src_min, const ARR<float> &src_max,
    const ARR<float> &dstn_min, const ARR<float> &dstn_max)
{
    ARR<float> interp_min;
    ARR<float> interp_max;
    const float ANIMSTEP = 10;

    interp_min.SetLen(dataset->dims);
    interp_max.SetLen(dataset->dims);

    int zero_mag = 0;
    int d;
    for (d=0; d<dataset->dims; d++){
        interp_min[d] = (src_min[d]-dstn_min[d])/ANIMSTEP;
        interp_max[d] = (dstn_max[d]-src_max[d])/ANIMSTEP;

        const float EPSILON = 0.0001;
        if ( fabs(interp_min[d])<EPSILON && fabs(interp_max[d])<EPSILON) zero_mag++;
    }

    // No need to magnify if no change
    //
    if (zero_mag==dataset->dims) return 0;

    // Set to the destination dimensions
    // first, do a quick datapoints marking,
    // then switch back to source dimensions
    //
    ARR<float> tmpmin(dataset->dims);
    ARR<float> tmpmax(dataset->dims);
    for (d=0; d<dataset->dims; d++) {
        tmpmin[d] = dataset->dim_min[d];
        tmpmax[d] = dataset->dim_max[d];
        dataset->dim_min[d] = dstn_min[d];
        dataset->dim_max[d] = dstn_max[d];
    }
    BrushOperMarkDataset();
    for (d=0; d<dataset->dims; d++) {
        dataset->dim_min[d] = tmpmin[d];
        dataset->dim_max[d] = tmpmax[d];
    }

    // Animate from source to destination
    //
    for (int i=0; i<ANIMSTEP; i++)
    {
        for (int d=0; d<dataset->dims; d++) {
            dataset->dim_min[d] -= interp_min[d];
            dataset->dim_max[d] += interp_max[d];
        }

        canvas->Refresh(true);
    }

    return 1;
}

//
// Magnify into the current brush region
//
void MagnifyBrushRegion(grCanvas *canvas)
{
    if (!dataset) return;

    // Compute brush extents
    //
    ARR<float> brush_min(dataset->dims);
    ARR<float> brush_max(dataset->dims);

    // Store the source extents
    for (int i=0; i<dataset->dims; i++) {
        brush_min[i] = curr_brush->Min(i);
        brush_max[i] = curr_brush->Max(i);
    }

    MAGNIFY = 1;
    MAGNIFY = AnimateMagnify(canvas,dataset->dim_min,dataset->dim_max,brush_min,brush_max);
}

//
// To de-magnify the current brush region,
// and bring the display back to its original
// state before *any* magnification takes place
//
void Demagnify(grCanvas *canvas)
{
    if (!dataset) return;

    MAGNIFY = 0;
//    AnimateMagnify(canvas,dataset->dim_min,dataset->dim_max,init_min,init_max);
    AnimateMagnify(canvas,dataset->dim_min,dataset->dim_max,dataset->org_dim_min,dataset->org_dim_max);
    BrushOperMarkDataset();
}
