////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef OGL_H
#define OGL_H

#include "Vec2.H"
#include "ARR.H"
#include "togl/togl.h"
#include "RGB.H"

struct grCanvas {

    public:
        RGBt fg_color;
        RGBt bg_color;
        bool xor;
        bool translucent;
        bool flatcolor;
        float alpha;
        Vec2 zoom;
        Vec2 translate;
        bool origin_at_lower_left;
        
        // Input state
        //
        bool button[3];
        Vec2 press_pos, curr_pos, prev_pos;
        static bool shiftkey, controlkey, altkey;
        static bool doublebuffer;

        // Event handlers
        //
        void (*display_proc)(grCanvas *canvas);
        void (*reshape_proc)(grCanvas *canvas);
        void (*motion_proc)(grCanvas *canvas, int x, int y);
        void (*buttonpress_proc)(grCanvas *canvas, int button, int x, int y);
        void (*buttonrelease_proc)(grCanvas *canvas, int button, int x, int y);
        void (*keypress_proc)(grCanvas *canvas, int key);
        void (*keyrelease_proc)(grCanvas *canvas, int key);

    protected:      
        // Canvas properties -- 
        // these values are queried/generated; DO NOT MODIFY. 
        //
        struct Togl *widget;
//        bool rgba, doublebuffer, cmap_installed;
//        int rsize, gsize, bsize, asize;
        GLbitfield clear_bits;
        GLuint font_base;

    public:
        grCanvas();
        struct Togl *GetWidget() { return widget; }

        static void grDisplayEvent( struct Togl *canvas_w );
        static void grReshapeEvent( struct Togl *canvas_w );
        static void grInitCanvas( struct Togl *canvas_w );
        static int grMotionEvent(struct Togl *togl,int argc,char *argv[]);
        static int grButtonPressEvent(struct Togl *togl,int argc,char *argv[]);
        static int grButtonReleaseEvent(struct Togl *togl,int argc,char *argv[]);
        static int grKeyPressEvent(struct Togl *togl,int argc,char *argv[]);
        static int grKeyReleaseEvent(struct Togl *togl,int argc,char *argv[]);
        static void grResizeMainCanvas( grCanvas *canvas );

        void ResetEventHandlers();      
        void Refresh(bool now = false);
        void MakeCurrent();
        void Clear();
        void Flush();
        void GetSize( int *w, int *h);
        void BeginDraw();
        void EndDraw();
        void SwapBuffer();
        void SetupTransformMatrix();
        void SetForeground( unsigned long color, float alpha=1.0 );
        void SetForeground( const RGBt &color, float alpha=1.0 );
        void SetBackground( unsigned long color );
        void SetXorMode( bool xor );
        void DrawLine( float x1, float y1, float x2, float y2);
        void DrawLines( const ARR<Vec2> &pts);
        void DrawRectangle( float x, float y, float width, float ht );
        void FillRectangle( float x, float y, float width, float ht);
        void DrawPoint( float x, float y, float size);
        void DrawPoints( const ARR<Vec2> &pts, float size);
        void DrawPolygon( const ARR<Vec2> &pts);
        void FillPolygon( const ARR<Vec2> &pts);
        void FillBand( const ARR<Vec2> &pts, float ratio);
        void DrawTransparentBand( const ARR<Vec2> &pts, float opacity_bottom, float opacity_top, RGBt color_bottom, RGBt color_top);
        void DrawTransparentBand2( const ARR<Vec2> &pts, float opacity_bottom, float opacity_top, float quartile);
//jing
		void DrawTransparentSquare2( const ARR<Vec2> &pts, float opacity_bottom, float opacity_top, float frac);
		void DrawTransparentSquare( const ARR<Vec2> &pts, float opacity_bottom, float opacity_top, RGBt color_bottom, RGBt color_top);
//jing
        void DrawString( const char *text, int num_chars=-1 );
        void DrawString( float x, float y, const char *text, int num_chars=-1);
        void DrawConfinedString( float x, float y, const char *text, float width);
        void StringExtents( const char *text, int num_chars, int *width, int *height );
        Vec2 GetScreenFontSize( const char *text, int n=-1 );
        Vec2 GetDataFontSize( const char *text, int n=-1 );
        void WriteImageSequence( const char *file_prefix );
        void WriteImage( const char *fname );
        Vec2 MapScreenToData( const Vec2 &sc );
        Vec2 MapDataToScreen( const Vec2 &p );
        void SetLowerLeftOrigin();
        void SetTopLeftOrigin();
        void SetHandCursor();
        void SetDefaultCursor();
        void SetCursor(const char *cursor_name);
#ifdef TOGL_X11
        bool PurgeEvent(int type, XEvent *ev=0);
#endif
};

extern struct Togl *grFindTogl(Tcl_Interp *interp, const char *name);
extern grCanvas *grFindCanvas(Tcl_Interp *interp, const char *name);

#endif
