////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "ogl.H"
#include "XmdvTool.H"
#include <stdlib.h>
#include <assert.h>
#include <GL/gl.h>

bool grCanvas::shiftkey = false;
bool grCanvas::controlkey = false;
bool grCanvas::altkey = false;
bool grCanvas::doublebuffer = true;

struct Togl *grFindTogl(Tcl_Interp *interp, const char *name)
{
    Tcl_CmdInfo infoPtr;
    if (Tcl_GetCommandInfo(interp,(char*)name,&infoPtr)==0) {
        return 0;
    }
    struct Togl *togl = (struct Togl*) infoPtr.clientData;
    return togl;
}

grCanvas *grFindCanvas(Tcl_Interp *interp, const char *name)
{
    struct Togl *togl = grFindTogl(interp,name);
    if (!togl) return 0;
    return (grCanvas*) Togl_GetClientData(togl);
}

//jing why drawbuffer twice here?
void grCanvas::grDisplayEvent(struct Togl *togl)
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    if (!doublebuffer) glDrawBuffer(GL_FRONT);

    canvas->Clear();
    canvas->SetupTransformMatrix();

    if (canvas->display_proc) canvas->display_proc(canvas);

#ifdef TOGL_X11
    canvas->PurgeEvent(Expose);
#endif
    if (doublebuffer) canvas->SwapBuffer();
    else glDrawBuffer(GL_BACK);
}

void grCanvas::grReshapeEvent(struct Togl *togl)
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int W, H;
    canvas->GetSize(&W,&H);

    glViewport( 0, 0, W, H );

    if (canvas->reshape_proc) canvas->reshape_proc(canvas);
}

int grCanvas::grMotionEvent(struct Togl *togl,int argc,char *argv[])
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int x = atoi(argv[2]);
    int y = Togl_Height( canvas->widget ) - atoi(argv[3]) - 1;

    canvas->prev_pos = canvas->curr_pos;
    canvas->curr_pos.X = x;
    canvas->curr_pos.Y = y;

    if (canvas->motion_proc) canvas->motion_proc(canvas,x,y);

    return TCL_OK;
}

int grCanvas::grButtonPressEvent(struct Togl *togl,int argc,char *argv[])
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int num = atoi(argv[2])-1;
    int x = atoi(argv[3]);
    int y = Togl_Height(canvas->widget) - atoi(argv[4]) - 1;

    if (altkey && num==0) {
        num =1;
    }

    canvas->button[num] = true;
    canvas->press_pos.X = canvas->curr_pos.X = x;
    canvas->press_pos.Y = canvas->curr_pos.Y = y;

    canvas->prev_pos = canvas->curr_pos;

    if (canvas->buttonpress_proc) canvas->buttonpress_proc(canvas,num,x,y);

    return TCL_OK;
}

int grCanvas::grButtonReleaseEvent(struct Togl *togl,int argc,char *argv[])
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int num = atoi(argv[2])-1;
    int x = atoi(argv[3]);
    int y = Togl_Height( canvas->widget ) - atoi(argv[4]) - 1;

    canvas->button[num] = false;
    canvas->press_pos.X = canvas->curr_pos.X = 0;
    canvas->press_pos.Y = canvas->curr_pos.Y = 0;

    canvas->prev_pos = canvas->curr_pos;

    if (canvas->buttonrelease_proc) canvas->buttonrelease_proc(canvas,num,x,y);

    return TCL_OK;
}

#include "X11/keysym.h"

int grCanvas::grKeyPressEvent(struct Togl *togl,int argc,char *argv[])
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int key = atoi(argv[2]);
    switch (key) {
        case XK_Shift_L:
        case XK_Shift_R:
            if (shiftkey) return TCL_OK;
            shiftkey = true;
            break;
        case XK_Control_L:
        case XK_Control_R:
            if (controlkey) return TCL_OK;
            controlkey = true;
            break;
        case XK_Alt_L:
        case XK_Alt_R:
            if (altkey) return TCL_OK;
            altkey = true;
            break;
    }

    // Grab all events if a modifier is pressed.
    // This will trap all key release events, even
    // when they are out of this canvas
    //
    if (shiftkey || altkey || controlkey) {
        Tk_Grab(Togl_Interp(togl),Togl_TkWin(togl),1);
    }

    if (canvas->keypress_proc) canvas->keypress_proc(canvas,key);

    return TCL_OK;
}

int grCanvas::grKeyReleaseEvent(struct Togl *togl,int argc,char *argv[])
{
    grCanvas *canvas = (grCanvas*) Togl_GetClientData(togl);
    assert(canvas);

    canvas->MakeCurrent();

    int key = atoi(argv[2]);
    switch (key) {
        case XK_Shift_L:
        case XK_Shift_R:
            shiftkey = false;
            break;
        case XK_Control_L:
        case XK_Control_R:
            controlkey = false;
            break;
        case XK_Alt_L:
        case XK_Alt_R:
            altkey = false;
            break;
    }

    // Release the grab we did in KeyPress
    if (!(shiftkey || altkey || controlkey)) {
        Tk_Ungrab(Togl_TkWin(togl));
    }

    if (canvas->keyrelease_proc) canvas->keyrelease_proc(canvas,key);

    return TCL_OK;
}

grCanvas::grCanvas() :
    zoom(1.0,1.0),
    translate(0,0),
    origin_at_lower_left(true),
    press_pos(0,0),
    curr_pos(0,0),
    display_proc(0),
    reshape_proc(0),
    motion_proc(0)
{
    button[0] = button[1] = button[2] = false;
}

void grCanvas::SetLowerLeftOrigin()
{
    origin_at_lower_left = true;
}

void grCanvas::SetTopLeftOrigin()
{
    origin_at_lower_left = false;
}

void grCanvas::ResetEventHandlers()
{
    display_proc = 0;
    reshape_proc = 0;
    motion_proc = 0;
    buttonpress_proc = 0;
    buttonrelease_proc = 0;
    keypress_proc = 0;
    keyrelease_proc = 0;
}

void grCanvas::grInitCanvas(struct Togl *togl)
{
    assert(togl);

    grCanvas *canvas = new grCanvas;
    assert(canvas);

    Togl_SetClientData(togl,canvas);

    canvas->widget = togl;

    canvas->MakeCurrent();

    glShadeModel(GL_SMOOTH);
    glDisable(GL_DEPTH_TEST);
//  glDisable(GL_DITHER);

    //
    // Setup some default values
    //
    canvas->fg_color = RGBt(interface_color[TEXT]);
    canvas->bg_color = RGBt(interface_color[BG]);
    canvas->xor = False;
    canvas->translucent = False;
    canvas->flatcolor = false;
    canvas->alpha = 1.0;

/*
    glGetIntegerv(GL_RED_BITS,&canvas->rsize);
    glGetIntegerv(GL_GREEN_BITS,&canvas->gsize);
    glGetIntegerv(GL_BLUE_BITS,&canvas->bsize);
    glGetIntegerv(GL_ALPHA_BITS,&canvas->asize);
    //DB(canvas->rsize VV canvas->gsize VV canvas->bsize VV canvas->asize);
*/
    //
    // Setup a default font to use
    //
#ifdef WIN32
	//jing
    const char *DEFAULT_FONT = "-adobe-helvetica-medium-r-normal--10-*-*-*-*-*-*-*";
#else
    const char *DEFAULT_FONT = "-adobe-helvetica-medium-r-normal--10-*-*-*-*-*-*-*";
#endif
    canvas->font_base = Togl_LoadBitmapFont(togl,DEFAULT_FONT);

    if (canvas->font_base==0) {
        printf("## Warning: OpenGL Font \"%s\" Initialization error.\n",DEFAULT_FONT);
        abort();
    }

    // If we have alpha bits, set it up to clear the
    // accum buffer too each time we clear screen.
    //
    canvas->clear_bits = GL_COLOR_BUFFER_BIT;
/*
    if (canvas->asize != 0) {
        canvas->clear_bits |= GL_ACCUM_BUFFER_BIT;
    }
*/
    // Create all static event handlers
    //
    Togl_CreateCommand( "Motion", grCanvas::grMotionEvent );
    Togl_CreateCommand( "ButtonPress", grCanvas::grButtonPressEvent );
//    Togl_CreateCommand( "DoubleClick", grCanvas::grDoubleClickEvent );
    Togl_CreateCommand( "ButtonRelease", grCanvas::grButtonReleaseEvent );
    Togl_CreateCommand( "KeyPress",grCanvas::grKeyPressEvent );
    Togl_CreateCommand( "KeyRelease",grCanvas::grKeyReleaseEvent );

    Tcl_CreateCommand(Togl_Interp(togl), Togl_Ident(togl), 0, togl, 0);

    canvas->Clear();
}

void grCanvas::Refresh(bool now)
{
    if (now) grDisplayEvent(widget);
    else Togl_PostRedisplay(widget);
}

void grCanvas::Flush()
{
    MakeCurrent();
    glFlush();
}

void grCanvas::GetSize(int *w, int *h)
{
    *w = Togl_Width(widget);
    *h = Togl_Height(widget);
}

void grCanvas::MakeCurrent()
{
    Togl_MakeCurrent(widget);
}

void grCanvas::SwapBuffer()
{
    Togl_SwapBuffers(widget);
    Flush();
}

void grCanvas::Clear()
{
    MakeCurrent();
    glClearColor(bg_color.R,bg_color.G,bg_color.B,0);
    glClear(clear_bits);
}

void grCanvas::SetupTransformMatrix()
{
    MakeCurrent();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glScalef(zoom.X,zoom.Y,1.0);
    glTranslatef(translate.X,translate.Y,1.0);
}

void grCanvas::SetXorMode( bool xor_is_on )
{
    xor = xor_is_on;
}

#include <sys/stat.h>

void grCanvas::WriteImageSequence( const char *file_prefix )
{
    char *default_prefix = "image.";
    if (file_prefix == 0) file_prefix = default_prefix;

 //   char fname[ strlen(file_prefix)+5 ];
//jing
	char* fname = new char[ strlen(file_prefix)+5 ];
//jing
    for (int seq=0;;seq++) {
        struct stat dummy;
        sprintf(fname,"%s%03d",file_prefix,seq);
        if (stat(fname,&dummy)<0) break;
    }

    WriteImage(fname);
//jing
	delete fname;
//jing
}

void grCanvas::WriteImage( const char *fname )
{
    Togl_PostRedisplay(GetWidget());
    Flush();

    FILE *fp = fopen(fname,"w");
    if (fp==0) {
        printf("Warning: file write error. Abort.\n");
        return;
    }

    int W, H;
    GetSize(&W,&H);

//    glReadBuffer(GL_FRONT); /* Need this? */

//    char rgb[4*W*H];
//jing
	unsigned char* rgb = new unsigned char[4*W*H];;
//jing
/*  glReadPixels(0,0,W,H,GL_RGBA,GL_UNSIGNED_BYTE,rgb);

    fprintf(fp,"P6 %d %d 255\n",W,H);
    for (int y=H-1;y>=0;y--) {
        int idx = y*(4*W);
        for (int x=0;x<W;x++,idx+=4) {
//          fwrite(&rgb[idx],sizeof(char),3,fp);
			rite(&rgb[idx+1],sizeof(char),1,fp);
			fwrite(&rgb[idx+2],sizeof(char),1,fp);
			fwrite(&rgb[idx],sizeof(char),1,fp);
        }
    }
*/
    glReadPixels(0,0,W,H,GL_RGBA,GL_UNSIGNED_BYTE,rgb);

    fprintf(fp,"P3 %d %d 255\n",W,H);
    for (int y=H-1;y>=0;y--) {
        int idx = y*(4*W);
        for (int x=0;x<W;x++,idx+=4) {
//          fwrite(&rgb[idx],sizeof(char),3,fp);
			fprintf(fp, "%d %d %d\n", (int)rgb[idx], (int)rgb[idx+1], (int)rgb[idx+2]);
        }
    }

    //printf("--- Wrote image of size %d x %d to file \"%s\"\n",W,H,fname);
    fclose(fp);
//jing
	delete rgb;
//jing
}

void grCanvas::SetForeground( const RGBt &color, float a )
{
    // Set normal foreground
    fg_color = color;
    alpha = a;
    if (a!=1.0) translucent = True;
    else translucent = False;
}

void grCanvas::SetForeground( unsigned long color, float a )
{
    SetForeground(RGBt(color),a);
}

void grCanvas::SetBackground( unsigned long color )
{
    // Set normal background
    bg_color = RGBt(color);
}

void grCanvas::grResizeMainCanvas( grCanvas *canvas )
{
    canvas->MakeCurrent();
//canvas->Clear();

    int W, H;
    canvas->GetSize(&W,&H);
    glViewport( 0, 0, W, H );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    const double BORD = 0.04;
    if (canvas->origin_at_lower_left) {
        glOrtho(-BORD,1+BORD,-BORD,1+BORD,-1,1);
    } else {
        glOrtho(-BORD,1+BORD,1+BORD,-BORD,-1,1);
    }
    glMatrixMode(GL_MODELVIEW);
}

void grCanvas::BeginDraw()
{
    MakeCurrent();

    if (xor == True)
    {
        // Setup XOR drawing mode
        glEnable(GL_COLOR_LOGIC_OP);
        glLogicOp(GL_INVERT);
        glDrawBuffer(GL_FRONT);
    } else {
        glColor4f(fg_color.R, fg_color.G, fg_color.B, alpha );
    }

    if (translucent == true) // && asize != 0)
    {
        // Turn on Translucency
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    }
/*
    if (flatcolor == False)  {
        // Turn on color interpolation
        glShadeModel(GL_SMOOTH);
    } else {
        glShadeModel(GL_FLAT);
    }
*/
}

void grCanvas::EndDraw()
{
    MakeCurrent();

    if (xor == True) {
        glDisable(GL_COLOR_LOGIC_OP);
        glDrawBuffer(GL_BACK);
    }
    if (translucent == True) { // && asize != 0)
        glDisable(GL_BLEND);
    }
/*
    if (flatcolor == False) {
        glShadeModel(GL_FLAT);
    }
*/
}

void grCanvas::DrawLine( float x1, float y1, float x2, float y2)
{
    BeginDraw();
    glBegin(GL_LINES);
    glVertex2f(x1,y1);
    glVertex2f(x2,y2);
    glEnd();
    EndDraw();
}

void grCanvas::DrawLines( const ARR<Vec2> &pts )
{
    if (pts.Len()<=0) return;
    BeginDraw();
    glBegin(GL_LINE_STRIP);
    glVertex2f(pts[0].X, pts[0].Y);
    for (int i=1; i<pts.Len(); i++) {
        glVertex2f(pts[i].X, pts[i].Y);
    }
    glEnd();
    EndDraw();
}

void grCanvas::DrawRectangle(float x, float y, float width, float ht )
{
    BeginDraw();
    glBegin(GL_LINE_LOOP);
    glVertex2f(x, y);
    glVertex2f(x+width, y);
    glVertex2f(x+width, y+ht);
    glVertex2f(x, y+ht);
    glEnd();
    EndDraw();
}

void grCanvas::FillRectangle( float x, float y, float width, float ht)
{
    BeginDraw();
    glBegin(GL_QUADS);
    glVertex2f(x, y);
    glVertex2f(x+width, y);
    glVertex2f(x+width, y+ht);
    glVertex2f(x, y+ht);
    glEnd();
    EndDraw();
}

void grCanvas::DrawPoint( float x, float y, float size)
{
    float half = 0.5*size;

    BeginDraw();
    glPointSize(size);
    glBegin(GL_POINTS);
    glVertex2f(x+half, y+half);
    glEnd();
    EndDraw();
}

void grCanvas::DrawPoints( const ARR<Vec2> &pts, float size)
{
    BeginDraw();
    glPointSize(size);
    glBegin(GL_POINTS);
    for (int i=0;i<pts.Len();i++) {
        glVertex2f(pts[i].X, pts[i].Y);
    }
    glEnd();
    EndDraw();
}

void grCanvas::DrawPolygon( const ARR<Vec2> &pts)
{
    BeginDraw();
    glBegin(GL_LINE_LOOP);
    for (int i=0;i<pts.Len();i++) {
        glVertex2f(pts[i].X, pts[i].Y);
    }
    glEnd();
    EndDraw();
}

void grCanvas::FillPolygon( const ARR<Vec2> &pts)
{
    BeginDraw();
    glBegin(GL_POLYGON);
    for (int i=0;i<pts.Len();i++) {
        glVertex2f(pts[i].X, pts[i].Y);
    }
    glEnd();
    EndDraw();
}

void grCanvas::FillBand( const ARR<Vec2> &pts, float ratio)
{
    assert(pts.Len()==4);
    assert(xor == False);

    Vec2 mid1, mid2;
    mid1 = (1-ratio)*pts[0] + (ratio)*pts[3];
    mid2 = (1-ratio)*pts[1] + (ratio)*pts[2];

    flatcolor = False;
    BeginDraw();
    glBegin(GL_POLYGON);
    glColor3f(bg_color.R, bg_color.G, bg_color.B);
    glVertex2f(pts[0].X, pts[0].Y);
    glVertex2f(pts[1].X, pts[1].Y);
    glColor3f(fg_color.R, fg_color.G, fg_color.B);
    glVertex2f(mid2.X, mid2.Y);
    glVertex2f(mid1.X, mid1.Y);
    glEnd();
    glBegin(GL_POLYGON);
    glColor3f(fg_color.R, fg_color.G, fg_color.B);
    glVertex2f(mid1.X, mid1.Y);
    glVertex2f(mid2.X, mid2.Y);
    glColor3f(bg_color.R, bg_color.G, bg_color.B);
    glVertex2f(pts[2].X, pts[2].Y);
    glVertex2f(pts[3].X, pts[3].Y);
    glEnd();
    EndDraw();
    flatcolor = True;
}

//#include "XmdvTool.h"
//extern unsigned long color_cells[NUM_COLOR_CELLS];

void grCanvas::DrawTransparentBand2( const ARR<Vec2> &pts, float opacity_bottom,
                        float opacity_top, float frac)
{
    Vec2 mid1 = (1-frac)*pts[0] + (frac)*pts[3];
    Vec2 mid2 = (1-frac)*pts[1] + (frac)*pts[2];

    ARR<Vec2> lo(4), hi(4);
    lo[0] = pts[0];
    lo[1] = pts[1];
    lo[2] = mid2;
    lo[3] = mid1;

    hi[0] = mid1;
    hi[1] = mid2;
    hi[2] = pts[2];
    hi[3] = pts[3];

//  RGBt highlight = RGBt(color_cells[CCELL_HIGHLIGHT]);
    RGBt highlight = RGBt(0x009900);

    if (frac < 0.5) {
        DrawTransparentBand( hi, opacity_bottom, opacity_top, fg_color, fg_color);
//      DrawTransparentBand( lo, opacity_bottom, opacity_bottom, highlight, highlight);
        DrawTransparentBand( lo, opacity_bottom, opacity_bottom, highlight, fg_color);
    } else {
        DrawTransparentBand( lo, opacity_bottom, opacity_top, fg_color, fg_color);
//      DrawTransparentBand( hi, opacity_top, opacity_top, highlight, highlight);
        DrawTransparentBand( hi, opacity_top, opacity_top, fg_color, highlight);
    }
}


void grCanvas::DrawTransparentBand( const ARR<Vec2> &pts,
                        float opacity_bottom, float opacity_top,
                        RGBt color_bottom, RGBt color_top)
{
    assert(pts.Len()==4);
    assert(xor == False);

    flatcolor = False;
    translucent = True;

    BeginDraw();
    glBegin(GL_QUAD_STRIP);
    glColor4f(color_bottom.R, color_bottom.G, color_bottom.B, alpha*opacity_bottom );
    glVertex2f(pts[1].X, pts[1].Y);
    glVertex2f(pts[0].X, pts[0].Y);

    glColor4f(color_top.R, color_top.G, color_top.B, alpha*opacity_top );
    glVertex2f(pts[2].X, pts[2].Y);
    glVertex2f(pts[3].X, pts[3].Y);
    glEnd();

    EndDraw();

    flatcolor = True;
    translucent = False;
}

//jing
void grCanvas::DrawTransparentSquare2( const ARR<Vec2> &pts, float opacity_bottom,
                        float opacity_top, float frac)
{
        DrawTransparentSquare( pts, opacity_top, opacity_bottom, fg_color, fg_color);
}

void grCanvas::DrawTransparentSquare( const ARR<Vec2> &pts,
                        float opacity_bottom, float opacity_top,
                        RGBt color_bottom, RGBt color_top)
{
    assert(pts.Len()==5);
    assert(xor == False);

    flatcolor = False;
    translucent = True;

    BeginDraw();
    glBegin(GL_TRIANGLES);

    glColor4f(color_bottom.R, color_bottom.G, color_bottom.B, alpha*opacity_bottom );
    glVertex2f(pts[0].X, pts[0].Y);
    glColor4f(color_top.R, color_top.G, color_top.B, alpha*opacity_top );
    glVertex2f(pts[1].X, pts[1].Y);
    glVertex2f(pts[2].X, pts[2].Y);

	glVertex2f(pts[2].X, pts[2].Y);
    glVertex2f(pts[3].X, pts[3].Y);
    glColor4f(color_bottom.R, color_bottom.G, color_bottom.B, alpha*opacity_bottom );
    glVertex2f(pts[0].X, pts[0].Y);
    
	
    glVertex2f(pts[0].X, pts[0].Y);
    glColor4f(color_top.R, color_top.G, color_top.B, alpha*opacity_top );
    glVertex2f(pts[3].X, pts[3].Y);
	glVertex2f(pts[4].X, pts[4].Y);

	
	glVertex2f(pts[4].X, pts[4].Y);
    glVertex2f(pts[1].X, pts[1].Y);
    glColor4f(color_bottom.R, color_bottom.G, color_bottom.B, alpha*opacity_bottom );
    glVertex2f(pts[0].X, pts[0].Y);
    glEnd();

    EndDraw();

    flatcolor = True;
    translucent = False;
}
//jing

void grCanvas::DrawString( const char *text, int N )
{
    assert(text);
    if (N<0) N=strlen(text);
    if (font_base == 0) return;
    glListBase(font_base);
    glCallLists(N,GL_UNSIGNED_BYTE,text);
}

void grCanvas::DrawString( float x, float y, const char *text, int N )
{
    assert(text);
    if (N<0) N=strlen(text);
    BeginDraw();
    glRasterPos2d(x,y);
    DrawString(text,N);
    EndDraw();
}

void grCanvas::StringExtents( const char *text, int num_chars, int *width, int *height )
{
    if (font_base == 0) return;

// HACK: We don't know how to
// query string extents in MS-Windows
//
*width = 5 * num_chars;
*height = 9;

/*
#ifdef TOGL_X11
    int dir,ascent,descent;
    XCharStruct overall;
    XTextExtents(fontinfo,text,num_chars,&dir,&ascent,&descent,&overall);
    if (height) *height = overall.ascent + overall.descent;
    if (width) *width = overall.width;
#endif
*/
}

Vec2 grCanvas::GetScreenFontSize( const char *text, int n )
{
    if (n<0) n = strlen(text);
    int font_width, font_height;
    StringExtents( text, n, &font_width, &font_height );
    return Vec2(font_width,font_height);
}

Vec2 grCanvas::GetDataFontSize( const char *text, int n)
{
    if (n<0) n = strlen(text);
    Vec2 screen_size = GetScreenFontSize(text,n);
    Vec2 data_font_size = MapScreenToData(screen_size) -
                          MapScreenToData(Vec2(0,0));
    return data_font_size;
}

void grCanvas::DrawConfinedString( float x, float y, const char *text, float width)
{
    int N = strlen(text);
    for (;N>0;N--) {
        Vec2 strsize = GetDataFontSize(text,N);
        if (strsize.X<width) {
            DrawString(x,y,text,N);
            return;
        }
    }
}

#ifdef TOGL_X11
bool grCanvas::PurgeEvent(int type, XEvent *ev)
{
    XEvent dummy;
    if (ev==0) ev=&dummy;
    bool purged = false;
    while ( XCheckTypedWindowEvent(Togl_Display(widget),Tk_WindowId(Togl_TkWin(widget)),type,ev) ) {
        purged=true;
    }
    return purged;
}
#endif

#include "GL/glu.h"

Vec2 grCanvas::MapScreenToData( const Vec2 &sc )
{
    GLdouble modelMatrix[16], projMatrix[16];
    GLint viewport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
    glGetIntegerv(GL_VIEWPORT,viewport);

    double wc[3];
    int ret = gluUnProject(sc.X,sc.Y,0,modelMatrix,projMatrix,viewport,wc,wc+1,wc+2);
    if (ret==GL_FALSE) {
        cerr << "gluProject failed.\n";
    }
    return Vec2(wc[0],wc[1]);
}

Vec2 grCanvas::MapDataToScreen( const Vec2 &p )
{
    GLdouble modelMatrix[16], projMatrix[16];
    GLint viewport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
    glGetIntegerv(GL_VIEWPORT,viewport);

    double sc[3];
    int ret = gluProject(p.X,p.Y,0,modelMatrix,projMatrix,viewport,sc,sc+1,sc+2);
    if (ret==GL_FALSE) {
        cerr << "gluProject failed.\n";
    }
    return Vec2(sc[0],sc[1]);
}

void grCanvas::SetHandCursor()
{
    #define hand_width 16
    #define hand_height 16
    #define hand_x_hot 3
    #define hand_y_hot 2

    static char hand_bits[] = {
       0x80, 0x01, 0x58, 0x0e, 0x64, 0x12, 0x64, 0x52, 0x48, 0xb2, 0x48, 0x92,
       0x16, 0x90, 0x19, 0x80, 0x11, 0x40, 0x02, 0x40, 0x04, 0x40, 0x04, 0x20,
       0x08, 0x20, 0x10, 0x10, 0x20, 0x10, 0x20, 0x10};

    static char hand_mask_bits[] = {
       0x80, 0x01, 0xd8, 0x0f, 0xfc, 0x1f, 0xfc, 0x5f, 0xf8, 0xff, 0xf8, 0xff,
       0xfe, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xfe, 0x7f, 0xfc, 0x7f, 0xfc, 0x3f,
       0xf8, 0x3f, 0xf0, 0x1f, 0xe0, 0x1f, 0xe0, 0x1f};

    Tk_Cursor cursor = Tk_GetCursorFromData(Togl_Interp(widget),Togl_TkWin(widget),
                        hand_bits,
                        hand_mask_bits,
                        hand_width, hand_height,
                        hand_x_hot, hand_y_hot,
                        Tk_GetUid("white"),
                        Tk_GetUid("black"));

    Tk_DefineCursor(Togl_TkWin(widget),cursor);
}

void grCanvas::SetDefaultCursor()
{
    SetCursor("left_ptr");
}

void grCanvas::SetCursor(const char *cursor_name)
{
    assert(cursor_name);
    Tk_Cursor cursor = Tk_GetCursor(Togl_Interp(widget),Togl_TkWin(widget),(char*)cursor_name);
    Tk_DefineCursor(Togl_TkWin(widget), cursor);
}
