////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef OKCDATA_H
#define OKCDATA_H


#include "ARR.H"
#include "RGB.H"

struct NominalField
{
    int range_min, range_max;
    ARR<char*> values;

    const char *MapValueToString(int what);
};


struct okcdata
{
    public:
        okcdata() :
            data_size(0),
            all_dims(0),
            dims(0)
        {}

        void Read(const char *fname);
        void Read2();
		void WriteBrushed(const char *fname);
        void GetData(ARR<float> &data, int line);
        void SetFieldOnOff(int i, bool state);
		void SetFieldOnOff2();
		void ReorderDimension(int which, int forward);
		void Placement_Sort_Glyph(int dimNumber, bool increase);
		void Placement_Sort_Space_Glyph(int dimNumber, bool increase);
		void Placement_Original_Space_Glyph();
		void Placement_Original_Glyph();
		void Placement_DataDriven_Glyph(int XdimNumber, int YdimNumber);
		void Placement_Derived_Glyph(int XdimNumber, int YdimNumber);
		void Placement_Shift_Glyph();

    public:
        // Original data read from file
        //
        char* datafile;
        int data_size;
        ARR<float> data_buf;
        int all_dims;
        ARR<float> all_dim_min, all_dim_max;
        ARR<float> cg_dim_min, cg_dim_max;
        ARR<int> all_cardinality;
        ARR<char*> all_names;
        ARR<NominalField*> all_nominal_fields;


        // Active dimensions in dataset
        //
        int dims;
        ARR<bool> on;
        ARR<float> dim_min, dim_max;
        ARR<int> cardinality;
        ARR<char*> names;
        ARR<NominalField*> nominal_fields;

        ARR<float> org_dim_min, org_dim_max;

        // Computed data -- Do not modify
        //
        float one_over_dims;

        enum mark_type { REGULAR, DISABLED, COLORED };
        ARR<mark_type> mark;
        ARR<RGBt> color;

        
		//used for reordering dimension
		ARR<int> map_file_to_buf;
        ARR<int> map_buf_to_file;

		//used for glyph placement
        float *basex, *basey;
		float blocksize, halfblocksize;
		int blockcnt;
};

#endif
