
#include "parcoordtool.H"
#include "parcoord.H"
#include "XmdvTool.H"
#include "ogl.H"
#include "brush.H"

static ARR<float> tmp_min, tmp_max;

static void SaveDimMinMax()
{
    // Save the current magnification
    // Setup up full scale magnification
    //
    tmp_min = dataset->dim_min;
    tmp_max = dataset->dim_max;
    dataset->dim_min = dataset->org_dim_min;
    dataset->dim_max = dataset->org_dim_max;
}

static void RestoreDimMinMax()
{
    // Restore magnification
    //
    dataset->dim_min = tmp_min;
    dataset->dim_max = tmp_max;
}

void ParcoordToolboxRedrawCanvas( grCanvas *canvas )
{
    if (!dataset) return;

    canvas->MakeCurrent();
    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    // Setup window parameters.
    // (duplicated from ogl.C)
    //
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    const double BORD = 0.1;
    glOrtho(-BORD,1+BORD,1+BORD,-BORD,-1,1);
    glMatrixMode(GL_MODELVIEW);

    SaveDimMinMax();

    // Paint the current brush
    DrawParcoordBrush(canvas,curr_brush);

    // Paint the axes and labels
    DrawParcoordAxes(canvas);

    RestoreDimMinMax();
}

void InputParcoordToolboxMouseMotion(grCanvas *canvas, int x, int y)
{
    SaveDimMinMax();
    InputParcoordMouseMotion(canvas,x,y);
    RestoreDimMinMax();
}

void InputParcoordToolboxButtonPress(grCanvas *canvas, int button, int x, int y)
{
    SaveDimMinMax();
    InputParcoordButtonPress(canvas,button,x,y);
    RestoreDimMinMax();
}

void InputParcoordToolboxButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    SaveDimMinMax();
    InputParcoordButtonRelease(canvas,button,x,y);
    RestoreDimMinMax();

    // Remark data after restoring min/max to sieve zoomed-out
    // points.
    appCanvasBrushHasChanged(Togl_Interp(canvas->GetWidget()));
}

void RegisterParcoordToolboxCanvas(grCanvas *canvas)
{
    assert(canvas);
    canvas->display_proc = ParcoordToolboxRedrawCanvas;
    canvas->motion_proc = InputParcoordToolboxMouseMotion;
    canvas->buttonpress_proc = InputParcoordToolboxButtonPress;
    canvas->buttonrelease_proc = InputParcoordToolboxButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
//    canvas->keyrelease_proc = appCanvasInputKeyRelease;
}
